# -*- coding: utf-8 -*-

"""A dictionary-like object for holding templates
"""

from collections.abc import Sequence
import logging

from .table import _Table
from .template import _Template

logger = logging.getLogger(__name__)


class _Templates(_Table):
    """Templates -- examples -- for subsets.

    A template serves two purposes. First, it provides a named
    instance for labeling similar subsets. The name of a template has
    two parts, a category and a name, e.g. residue/ala,
    molecule/water, or general/group1. This provides a way to group
    and identify subsets.

    Second, the template may optionally point to a configuration of a
    system, in which case the atoms in the template configuration are
    templates for the atoms in subsets created for the template. This
    provides a way to create new molecules or fragments from a
    template, or to transfer information from a template into a
    configuration. For example, this could be used to transfer
    canonical names and atom categories residue templates onto a
    protein molecule.

    This class handles the templates.

    :meta public:
    """

    def __init__(self, system_db, logger=logger):
        """Initialize the instance."""
        super().__init__(system_db, table="template", logger=logger)

    # def __repr__(self):
    #     """The string representation of this object"""
    #     raise NotImplementedError()

    # def __str__(self):
    #     """The pretty string representation of this object"""
    #     raise NotImplementedError()

    @property
    def categories(self):
        """Return the current categories of templates."""
        sql = "SELECT DISTINCT category FROM template ORDER BY category ASC"
        return [x[0] for x in self.db.execute(sql)]

    @property
    def n_templates(self):
        """The number of templates."""
        return self.n_rows

    def create(
        self, name, category="general", configuration=None, smiles=True, **kwargs
    ):
        """Create one templates.

        This is an explicit wrapper of the Table method append, specializing
        it for creating one template.

        Parameters
        ----------
        name : str
            The name of the template/subset. The name-category pair must be
            unique.
        category : str = 'general'
            An optional category of template/subset, e.g. 'general', 'residue',
            'molecule'
        configuration : int or _Configuration = None
            An optional template configuration or its id to use.
        smiles : bool = True
            Create and store the canonical SMILES for full templates.
        kwargs : keyword arguments
            Other existing attributes and values.

        Returns
        -------
        _template
            The created template.
        """
        if configuration is None:
            cid = None
        else:
            if isinstance(configuration, int):
                cid = configuration
                configuration = self.system_db.get_configuration(cid)
            else:
                cid = configuration.id
            canonical_smiles = configuration.to_smiles(canonical=True)
            kwargs["canonical_smiles"] = canonical_smiles

        tid = self.append(category=category, name=name, configuration=cid, **kwargs)[0]
        return _Template(self.system_db, tid)

    def create_many(
        self, name, category="general", configuration=None, smiles=True, **kwargs
    ):
        """Create one or more templates.

        This is an explicit wrapper of the Table method append, specializing
        it for templates.

        Parameters
        ----------
        name : str
            The name of the template/subset. The name-category pair must be
            unique.
        category : str = 'general'
            An optional category of template/subset, e.g. 'general', 'residue',
            'molecule'
        configuration : int or _Configuration = None
            An optional template configuration or its id to use.
        smiles : bool = True
            Create and store the canonical SMILES for full templates.
        kwargs : keyword arguments
            Other existing attributes and values.

        Returns
        -------
        [_Template]
            The created templates.
        """
        if configuration is None:
            cids = None
        else:
            cids = []
            if isinstance(configuration, Sequence):
                kwargs["canonical_smiles"] = []
                if isinstance(configuration, int):
                    cid = configuration
                    configuration = self.system_db.get_configuration(cid)
                else:
                    cid = configuration.id
                canonical_smiles = configuration.to_smiles(canonical=True)
                kwargs["canonical_smiles"].append(canonical_smiles)
                cids.append[cid]
            else:
                if isinstance(configuration, int):
                    cids = configuration
                    configuration = self.system_db.get_configuration(cids)
                else:
                    cids = configuration.id
                canonical_smiles = configuration.to_smiles(canonical=True)
                kwargs["canonical_smiles"] = canonical_smiles

        tids = self.append(category=category, name=name, configuration=cids, **kwargs)
        return [_Template(self.system_db, tid) for tid in tids]

    def exists(self, name, category="general"):
        """Return whether a given template exists.

        Parameters
        ----------
        name : str
            The name of the template/subset. The name-category pair must be
            unique.
        category : str = 'general'
            An optional category of template/subset, e.g. 'general', 'residue',
            'molecule'

        Returns
        -------
        bool
            Whether the template exists.
        """
        sql = "SELECT COUNT(*) FROM template WHERE category = ? AND name = ?"
        self.cursor.execute(sql, (category, name))
        return self.cursor.fetchone()[0]

    def get(self, name, category="general"):
        """Return a template.

        Parameters
        ----------
        name : str
            The id or name of the template/subset. If the id is given, the
            category is ignored. Otherwise the template with the given name
            and category is returned.
        category : str = 'general'
            An optional category of template/subset, e.g. 'general', 'residue',
            'molecule'

        Returns
        -------
        _Template
            The template, or None if it does not exist.
        """
        if isinstance(name, str):
            # A name is given
            tid = self.get_id(name, category=category)
            if tid is None:
                return None
        else:
            # The id given
            tid = name
        return _Template(self.system_db, tid)

    def get_id(self, name, category="general"):
        """Return the id of a template.

        Parameters
        ----------
        name : str
            The name of the template/subset. The name-category pair must be
            unique.
        category : str = 'general'
            An optional category of template/subset, e.g. 'general', 'residue',
            'molecule'

        Returns
        -------
        int
            The id of the template, or None if it does not exist.
        """
        sql = "SELECT id FROM template WHERE category = ? AND name = ?"
        self.cursor.execute(sql, (category, name))
        result = self.cursor.fetchone()
        if result is not None:
            result = result[0]

        return result

    def names(self, category):
        """Return the current names of templates of the given category.

        Parameters
        ----------
        category : str
            The category of template/subset, e.g. 'general', 'residue',
            'molecule'

        Returns
        -------
        [str]
            The names of templates of the given category.
        """
        sql = "SELECT name FROM template WHERE category = ? ORDER BY name ASC"
        return [x[0] for x in self.db.execute(sql, (category,))]



# -*- coding: utf-8 -*-

import collections.abc
import itertools
import logging

from molsystem.frozencolumn import _FrozenColumn as FrozenColumn

logger = logging.getLogger(__name__)


class _Column(FrozenColumn, collections.abc.MutableSequence):
    """A list-like object for holding a column data

    This is a wrapper around a single column in a SQL table. Setting a value
    updates the SQL database appropriately.
    """

    def __init__(self, table, column: str, sql=None, where=""):
        self._rowids = None
        super().__init__(table, column, sql, where)

    def __setitem__(self, index, value) -> None:
        """Allow x[index] access to the data"""
        length = self._table.length_of_values(value)
        db = self._table.db
        table = self._table.table
        if isinstance(index, int):
            if length == 0:
                # Scalar
                db.execute(
                    f"UPDATE {table} SET {self.column} = ? WHERE rowid = ?",
                    (value, self._rowids[index]),
                )
                self._data[index] = value
            elif length == 1:
                # One-element list
                db.execute(
                    f"UPDATE {table} SET {self.column} = ? WHERE rowid = ?",
                    (value[0], self._rowids[index]),
                )
                self._data[index] = value[0]
            else:
                raise IndexError("Only 1 value required to update 1 item")
        elif isinstance(index, slice):
            rowids = []
            for rowid in itertools.islice(
                self._rowids, index.start, index.stop, index.step
            ):
                rowids.append(rowid)
            parameters = []
            if length == 0:
                for rowid in rowids:
                    parameters.append((value, rowid))
                self._data[index] = [value] * len(rowids)
            elif length == 1:
                for rowid in rowids:
                    parameters.append((value[0], rowid))
                self._data[index] = [value[0]] * len(rowids)
            elif length == len(rowids):
                for rowid, val in zip(rowids, value):
                    parameters.append((val, rowid))
                self._data[index] = value
            else:
                raise IndexError(
                    f"The number of values ({length}) must be 1 or the size "
                    f"of the slice ({len(rowids)})."
                )
            db.executemany(
                f"UPDATE {table} SET {self.column} = ? WHERE rowid = ?", parameters
            )
        db.commit()

    def __delitem__(self, index, value) -> None:
        """Do NOT allow deletion!"""
        raise RuntimeError("Items may not be removed from a Column")

    def __repr__(self) -> str:
        """The string representation of this object"""
        return repr([(r, x) for r, x in zip(self._rowids, self._data)])

    def insert(self, index, value) -> None:
        """Do NOT allow insertion!"""
        raise RuntimeError("Items may not be inserted into a Column")

    def _initialize(self):
        """Get the column data from the underlying table and cache it."""
        self._data = []
        self._rowids = []
        if self._sql is None:
            sql = (
                f"SELECT rowid, {self.column}"
                f"  FROM {self._table.table} {self._where}"
            )
        else:
            sql = self._sql

        for row in self._table.db.execute(sql):
            self._rowids.append(row[0])
            self._data.append(row[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `molsystem` package.

These tests center on angiotensin, a small peptide.
"""

from pathlib import Path
import pprint  # noqa: F401
import pytest  # noqa: F401

from molsystem import SystemDB

path = Path(__file__).resolve().parent
data_path = path / "data"

sequence = "H-Asp-Arg-Val-Tyr-Ile-His-Pro-Phe-OH"

# SMILES for angiotensin II
# H-Asp-Arg-Val-Tyr-Ile-His-Pro-Phe-OH
SMILES = (
    "[NH2][C@@H](CC(=O)O)C(=O)"
    "[NH][C@@H](CCCNC(=[NH2])N)C(=O)"
    "[NH][C@@H](C(C)C)C(=O)"
    "[NH][C@@H](Cc1ccc(cc1)O)C(=O)"
    "[NH][C@@H]([C@@H](CC)C)C(=O)"
    "[NH][C@@H](CC1=CNC=[NH]1)C(=O)"
    "[N]1[C@@H](CCC1)C(=O)"
    "[NH][C@@H](Cc1ccccc1)C(=O)O"
)
SMILES = (
    "[NH2][C@@H](CC(=O)O)C(=O)"
    "[NH][C@@H](CCCNC(=N)N)C(=O)"
    "[NH][C@@H](C(C)C)C(=O)"
    "[NH][C@@H](Cc1ccc(cc1)O)C(=O)"
    "[NH][C@@H]([C@@H](CC)C)C(=O)"
    "[NH][C@@H](CC1=CNC=N1)C(=O)"
    "[N]1[C@@H](CCC1)C(=O)"
    "[NH][C@@H](Cc1ccccc1)C(=O)O"
)
sidechains = {
    "ALA_LL": "[CH3X4]",
    "ARG_LL": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3]([NH2X3])=[NH2X3]",
    "ARG_LL_DHH12": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3](=[NH1X2])[NH2X3]",
    "ARG_LL_DHH22": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3]([NH2X3])=[NH1X2]",
    "ARG_LL_RNH1": "[CH2X4][CH2X4][CH2X4][NH1X3][CH0X3](=[NH2X3])[NH2X3]",
    "ASN_LL": "[CH2X4][CH0X3](=[OH0X1])[NH2X3]",
    "ASP_LL": "[CH2X4][CH0X3](=[OH0X1])[OH1X2]",
    "ASP_LL_DHD2": "[CH2X4][CH0X3](=[OH0X1])[OH0X1]",
    "CYS_LL": "[CH2X4][SH1X2]",
    "CYS_LL_DHG": "[CH2X4][SH0X1]",
    "GLN_LL": "[CH2X4][CH2X4][CH0X3](=[OH0X1])[NH2X3]",
    "GLU_LL": "[CH2X4][CH2X4][CH0X3](=[OH0X1])[OH1X2]",
    "GLU_LL_DHE2": "[CH2X4][CH2X4][CH0X3](=[OH0X1])[OH0X1]",
    "HIS_LL": "[CH2X4][cH0X3][nH1X3][cH1X3][nH1X3][cH1X3]",
    "HIS_LL_DHD1": "[CH2X4][cH0X3][nH0X2][cH1X3][nH1X3][cH1X3]",
    "HIS_LL_DHE2": "[CH2X4][cH0X3][nH1X3][cH1X3][nH0X2][cH1X3]",
    "ILE_LL": "[CH1X4]([CH2X4][CH3X4])[CH3X4]",
    "LEU_LL": "[CH2X4][CH1X4]([CH3X4])[CH3X4]",
    "LYS_LL": "[CH2X4][CH2X4][CH2X4][CH2X4][NH3X4]",
    "LYS_LL_DHZ3": "[CH2X4][CH2X4][CH2X4][CH2X4][NH2X3]",
    "MET_LL": "[CH2X4][CH2X4][SH0X2][CH3X4]",
    "PHE_LL": "[CH2X4][cH0X3][cH1X3][cH1X3][cH1X3][cH1X3][cH1X3]",
    "SER_LL": "[CH2X4][OH1X2]",
    "SER_LL_DHG": "[CH2X4][OH0X1]",
    "THR_LL": "[CH1X4]([OH1X2])[CH3X4]",
    "THR_LL_DHG1": "[CH1X4]([OH0X1])[CH3X4]",
    "TRP_LL": (
        "[CH2X4][CH0X3]=[CH1X3][NH1X3][CH0X3]=[CH0X3][CH1X3]=[CH1X3]" "[CH1X3]=[CH1X3]"
    ),
    "TRP_LL_DHE1": (
        "[CH2X4][CH0X3]=[CH1X3][NH0X2][CH0X3]=[CH0X3][CH1X3]=[CH1X3]" "[CH1X3]=[CH1X3]"
    ),
    "TYR_LL": "[CH2X4][cH0X3]1[cH1X3][cH1X3][cH0X3]([OH1X2])[cH1X3][cH1X3]1",
    "TYR_LL_DHH": ("[CH2X4][CH0X3]=[CH1X3][CH1X3]=[CH0X3]([CH1X3]=[CH1X3])[OH0X1]"),
    "VAL_LL": "[CH1X4]([CH3X4])[CH3X4]",
}

full_smarts = {
    "GLY_LL": "[NH1X2][CH2X4][CH0X2]=[OH0X1]",
    "PRO_LL": "[NH0X2][CH1X4]([CH0X2]=[OH0X1])[CH2X4][CH2X4][CH2X4]",
}


@pytest.fixture(scope="module")
def testdb():
    """Create a database with amino acids and templates for each."""
    db = SystemDB(filename="file:aa_templates_db?mode=memory&cache=shared")
    db.read_cif_file(data_path / "aa-variants-v1.cif")

    # Make the templates
    templates = db.templates
    for system in db.systems:
        tmp = system.name.split("_")
        if len(tmp) == 1:
            category = "amino acid"
        else:
            if tmp[1] == "LL":
                category = "residue"
            elif tmp[1] == "LEO2":
                category = "C-terminal residue"
            elif tmp[1] == "LEO2H":
                category = "protonated C-terminal residue"
            elif tmp[1] == "LFOH":
                category = "amino acid free neutral"
            elif tmp[1] == "LFZW":
                category = "amino acid free zwitterion"
            elif tmp[1] == "LSN3":
                category = "protonated N-terminal residue"
            else:
                raise ValueError(f"Don't recognize {tmp[1]} in {system.name}")

        templates.create(
            name=system.name, category=category, configuration=system.configuration.id
        )

    # Angiotensin II from SMILES
    system = db.create_system("angiotensin")
    configuration = system.create_configuration("SMILES")
    configuration.from_smiles(SMILES)

    # And read in
    system.read_cif_file(data_path / "1n9v.cif")

    yield db

    db.close()


@pytest.fixture()
def angiotensin(configuration):
    """A system with one configuration, containing angiotensin."""
    configuration.from_smiles(SMILES)

    return configuration


def test_correctness(angiotensin):
    """Check that the structure is correct."""
    assert angiotensin.formula[0] == "C50 H71 N13 O12"
    assert abs(angiotensin.mass - 1046.2) < 0.1


def test_testdb(testdb):
    """Test that the test database is correct."""
    db = testdb
    # There are 218 structures in aa-variants, plus the 1 angiotensin system
    assert db.n_systems == 219

    templates = db.templates
    assert templates.n_templates == 218

    system = db.get_system("angiotensin")
    # There is the SMILES configuration and 21 from the ensemble
    assert system.n_configurations == 22

    smiles = system.get_configuration("SMILES")

    assert smiles.atoms.n_atoms == 146
    assert smiles.formula[0] == "C50 H71 N13 O12"
    assert abs(smiles.mass - 1046.2) < 0.1

    angiotensin = system.get_configuration("representative")
    assert angiotensin.atoms.n_atoms == 146
    assert angiotensin.formula[0] == "C50 H71 N13 O12"
    assert abs(angiotensin.mass - 1046.2) < 0.1


def test_all_residue_search(testdb):
    """Testing locating residues in a peptide."""
    residues = {
        "ARG_LL_DHH12": [(9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19)],
        "ARG_LL_DHH22": [(9, 10, 11, 12, 13, 14, 15, 17, 16, 18, 19)],
        "HIS_LL_DHD1": [(47, 48, 49, 50, 54, 53, 52, 51, 55, 56)],
        "ILE_LL": [(39, 40, 41, 42, 43, 44, 45, 46)],
        "PHE_LL": [(64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74)],
        "TYR_LL": [(27, 28, 29, 30, 31, 32, 33, 36, 34, 35, 37, 38)],
        "VAL_LL": [(20, 21, 22, 23, 24, 25, 26)],
    }
    n_terminal = {
        "ASP_LL": [(1, 2, 3, 4, 5, 6, 7, 8)],
    }
    c_terminal = {
        "PHE_LL": [(64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75)],
    }

    db = testdb

    if False:
        print("SMARTS")
        for name in db.templates.names("residue"):
            tmp = db.templates.get(name, "residue")
            print(f"'{name}': '{tmp.smarts}',")

        print("SMILES")
        for name in db.templates.names("residue"):
            tmp = db.templates.get(name, "residue")
            print(f"'{name}': '{tmp.smiles}',")

    system = db.get_system("angiotensin")
    angiotensin = system.get_configuration("SMILES")

    # print(angiotensin.smiles)

    for name, sc in sidechains.items():
        smarts = f"[NH1X3][C@@H]({sc})[CX3]=[OX1]"
        result = angiotensin.find_substructures(smarts)
        if len(result) == 0:
            assert name not in residues
        else:
            if name not in residues or result != residues[name]:
                print(f"'{name}': {result},")
            assert result == residues[name]

    for name, sc in sidechains.items():
        smarts = f"[NH2][C@@H]({sc})[CX3]=[OX1]"
        result = angiotensin.find_substructures(smarts)
        if len(result) == 0:
            assert name not in n_terminal
        else:
            if name not in n_terminal or result != n_terminal[name]:
                print(f"'{name}': {result},")
            assert result == n_terminal[name]

    for name, sc in sidechains.items():
        smarts = f"[NX3][C@@H]({sc})[CX3](=[OX1])[OH1X2]"
        result = angiotensin.find_substructures(smarts)
        if len(result) == 0:
            assert name not in c_terminal
        else:
            if name not in c_terminal or result != c_terminal[name]:
                print(f"'{name}': {result},")
            assert result == c_terminal[name]



# -*- coding: utf-8 -*-

"""A dictionary-like object for holding bonds

Based on tables in an SQLite database.
"""

from collections.abc import Sequence
from itertools import zip_longest
import logging
import sqlite3

import numpy as np
import pandas

from .column import _Column
from .table import _Table
from .frozencolumn import _FrozenColumn

logger = logging.getLogger(__name__)


def grouped(iterable, n):
    "s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,...s3n-1), ..."
    return zip_longest(*[iter(iterable)] * n)


class _Bonds(_Table):
    """The Bonds class describes the bonds in the system.

    :meta public:
    """

    def __init__(self, configuration):

        self._configuration = configuration

        self._system = None

        self._bondset = self._configuration.bondset

        super().__init__(configuration._system_db, "bond")

    def __enter__(self):
        """Copy the tables to a backup for a 'with' statement."""
        self.system_db["bondset_bond"].__enter__()
        self.system_db["bond"].__enter__()
        return self

    def __exit__(self, etype, value, traceback):
        """Handle returning from a 'with' statement."""
        if etype is None:
            self.configuration.version = self.configuration.version + 1
        self.system_db["bondset_bond"].__exit__(etype, value, traceback)
        return self.system_db["bond"].__exit__(etype, value, traceback)

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        raise NotImplementedError()

    @property
    def ids(self):
        """The ids of the bonds."""
        return self.get_ids()

    @property
    def bondorders(self):
        """The bond orders."""
        return self.get_column_data("bondorder")

    @property
    def bondset(self):
        """The bondset for these bonds."""
        return self._bondset

    @property
    def configuration(self):
        """Return the configuration."""
        return self._configuration

    @property
    def cursor(self):
        """The database connection."""
        return self.system_db.cursor

    @property
    def db(self):
        """The database connection."""
        return self.system_db.db

    @property
    def n_bonds(self):
        """The number of bonds."""
        self.cursor.execute(
            "SELECT COUNT(*) FROM bondset_bond WHERE bondset = ?", (self.bondset,)
        )
        return self.cursor.fetchone()[0]

    @property
    def system_db(self):
        """The system database that we belong to."""
        return self._system_db

    @property
    def system(self):
        """The system that we belong to."""
        if self._system is None:
            self.cursor.execute(
                "SELECT system FROM configuration WHERE id = ?", (self.id,)
            )
            self._system = self.system_db.get_system(self.cursor.fetchone()[0])
        return self._system

    def append(self, **kwargs):
        """Append one or more bonds

        The keys give the field for the data. If an existing field is not
        mentioned, then the default value is used, unless the default is None,
        in which case an error is thrown. It is an error if there is not a
        field corrresponding to a key.

        Parameters
        ----------
        bonds : [sqlite3.Rows] or {str : [int]}
            The bonds as either SQLite Rows or a dictionary of values or lists.
            Any other arguments override the same value in 'bonds'.
        i : [int]
            The atom indices of the first atom.
        j : [int]
            The atom indices of the second atom.
        bondorder : [int] optional
            Bond orders, defaults to 1 for single bonds.
        other : [int, float, str] optional
            Other optional attributes of bonds.
        """
        if "bonds" in kwargs:
            bonds = kwargs.pop("bonds")
            if isinstance(bonds, sqlite3.Row):
                for item in bonds.keys():
                    if item != "id":
                        kwargs[item] = bonds[item]
            elif isinstance(bonds, Sequence) and isinstance(bonds[0], sqlite3.Row):
                for item in bonds[0].keys():
                    if item != "id":
                        kwargs[item] = [row[item] for row in bonds]
            else:
                try:
                    for key, value in bonds.items():
                        if isinstance(value, Sequence):
                            kwargs[key] = value
                        else:
                            kwargs[key] = [value]
                except AttributeError:
                    raise TypeError(
                        "bonds argument must be sqlite3.Row or a mapping "
                        f"(dict) type, not {type(bonds)}."
                    )

        # Check keys and lengths of added bonds
        if "i" not in kwargs or "j" not in kwargs:
            raise KeyError("The atoms i & j are required!")

        i = kwargs.pop("i")
        j = kwargs.pop("j")
        try:
            len_i = len(i)
        except TypeError:
            len_i = 1
            i = (i,)
        try:
            len_j = len(j)
        except TypeError:
            len_j = 1
            j = (j,)

        if len_i == 0 and len_j == 0:
            return

        if len_i == 1:
            if len_j > 1:
                i = len_j * [i[0]]
        elif len_j == 1:
            j = len_i * [j[0]]
        elif len_i != len_j:
            raise IndexError(
                f'key "j" has the wrong number of values, {len_j}. '
                f"Should be 1 or the number of values in i, {len_i}."
            )
        # Ensure that i < j

        if isinstance(i, np.ndarray):
            i = i.tolist()
        if isinstance(j, np.ndarray):
            j = j.tolist()

        i2 = []
        j2 = []
        for i_, j_ in zip(i, j):
            # will need to handle offsets here at some point
            if not isinstance(i_, int) or not isinstance(j_, int):
                raise TypeError("'i' and 'j', the atom indices, must be integers")
            if i_ < j_:
                i2.append(i_)
                j2.append(j_)
            else:
                i2.append(j_)
                j2.append(i_)

        ids = super().append(i=i2, j=j2, **kwargs)

        # And to the bondset
        table = _Table(self.system_db, "bondset_bond")
        table.append(bondset=self.bondset, bond=ids)

        return ids

    def bonds(self, *args):
        """Returns an iterator over the bonds.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        sqlite3.Cursor
            A cursor that returns sqlite3.Row objects for the bonds.
        """
        sql = (
            "SELECT bond.* FROM bond, bondset_bond"
            " WHERE bond.id = bondset_bond.bond"
            "   AND bondset_bond.bondset = ?"
        )
        return self.db.execute(sql, (self.bondset,))

    def contains_bond(self, i, j):
        """Whether there is a bond between atoms i and j."""
        # get canonical order
        if i > j:
            j, i = i, j
        sql = (
            "SELECT COUNT(*) FROM bond, bondset_bond"
            " WHERE bond.i = ? AND bond.j = ?"
            "   AND bondset_bond.bondset = ?"
        )
        self.cursor.execute(sql, (i, j, self.bondset))
        return self.cursor.fetchone()[0] == 1

    def delete_bond(self, i, j=None, force=False):
        """Delete a bond, if present.

        Parameters
        ----------
        i : int
            Either the first atom in the bond or the bond id.
        j : int = None
            If not None, then the second atom in the bond.
        force : bool = False
            If true, ignore missing bonds.
        """
        # get canonical order
        if j is None:
            bond_id = i
        else:
            if i > j:
                j, i = i, j
            sql = (
                "SELECT bond.id FROM bond, bondset_bond"
                " WHERE bond.i = ? AND bond.j = ?"
                "   AND bondset_bond.bondset = ?"
            )
            self.cursor.execute(sql, (i, j, self.bondset))
            bond_id = self.cursor.fetchone()
            if bond_id is not None:
                bond_id = bond_id[0]

        if bond_id is None:
            if not force:
                raise ValueError("The bond did not exist.")
        else:
            self.cursor.execute("DELETE FROM bond WHERE id = ?", (bond_id,))

    def delete(self, atoms=None):
        """Deletes all the bonds, optionally from atoms.

        Parameters
        ----------
        atoms : [int] = None
            The list of atoms whose bonds are deleted

        Returns
        -------
        None
        """
        if atoms is not None:
            sql = (
                "DELETE FROM bond"
                " WHERE (i = ? OR j = ?)"
                "   AND id in ("
                "              SELECT bond FROM bondset_bond"
                "              WHERE bondset = ?"
                "             )"
            )
            parameters = [(i, i, self.bondset) for i in atoms]
            self.db.executemany(sql, parameters)
        else:
            sql = (
                "DELETE FROM bond"
                " WHERE id in ("
                "              SELECT bond FROM bondset_bond"
                "               WHERE bondset = ?"
                "             )"
            )
            self.db.execute(sql, (self.bondset,))

    def diff(self, other):
        """Difference between these bonds and another

        Parameters
        ----------
        other : _Bonds
            The other bonds to diff against

        Result
        ------
        result : Dict
            The differences, described in a dictionary
        """
        result = {}

        # Check the columns
        columns = self._columns()
        other_columns = other._columns()

        column_defs = ", ".join(columns)
        other_column_defs = ", ".join(other_columns)

        if columns == other_columns:
            column_def = column_defs
        else:
            added = columns - other_columns
            if len(added) > 0:
                result["columns added"] = list(added)
            deleted = other_columns - columns
            if len(deleted) > 0:
                result["columns deleted"] = list(deleted)

            in_common = other_columns & columns
            if len(in_common) > 0:
                column_def = ", ".join(in_common)
            else:
                # No columns shared
                return result

        # Need to check the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        db = self.system_db
        other_db = other.system_db

        detach = False
        schema = self.schema
        if db.filename != other_db.filename:
            if db.is_attached(other_db):
                other_schema = db.attached_as(other_db)
            else:
                # Attach the other system_db in order to do comparisons.
                other_schema = self.system_db.attach(other_db)
                detach = True
        else:
            other_schema = other.schema

        bondset = self.bondset
        other_bondset = other.bondset

        changed = {}
        last = None
        sql = f"""
        SELECT * FROM
        (
          SELECT {column_def}
            FROM {other_schema}.bond
           WHERE id
              IN (
                  SELECT bond
                    FROM {other_schema}.bondset_bond
                   WHERE bondset = {other_bondset}
                  )
          EXCEPT
          SELECT {column_def}
            FROM {schema}.bond
           WHERE id
              IN (
                  SELECT bond
                    FROM {schema}.bondset_bond
                    WHERE bondset = {bondset}
                 )
        )
         UNION ALL
        SELECT * FROM
        (
          SELECT {column_def}
            FROM {schema}.bond
           WHERE id
              IN (
                  SELECT bond
                    FROM {schema}.bondset_bond
                   WHERE bondset = {bondset}
                 )
          EXCEPT
          SELECT {column_def}
            FROM {other_schema}.bond
           WHERE id
              IN (
                  SELECT bond
                    FROM {other_schema}.bondset_bond
                   WHERE bondset = {other_bondset}
                 )
        )
        ORDER BY id
        """

        for row in self.db.execute(sql):
            if last is None:
                last = row
            elif row["id"] == last["id"]:
                # changes = []
                changes = set()
                for k1, v1, v2 in zip(last.keys(), last, row):
                    if v1 != v2:
                        changes.add((k1, v1, v2))
                changed[row["id"]] = changes
                last = None
            else:
                last = row
        if len(changed) > 0:
            result["changed"] = changed

        # See about the rows added
        added = {}
        sql = f"""
        SELECT {column_defs}
          FROM {schema}.bond
         WHERE id
            IN (
                SELECT bond
                  FROM {schema}.bondset_bond
                 WHERE bondset = {bondset}
               )
           AND id
        NOT IN (
                SELECT bond
                  FROM {other_schema}.bondset_bond
                 WHERE bondset = {other_bondset}
               )
        """
        for row in self.db.execute(sql):
            added[row["id"]] = row[1:]

        if len(added) > 0:
            result["columns in added rows"] = row.keys()[1:]
            result["added"] = added

        # See about the rows deleted
        deleted = {}
        sql = f"""
        SELECT {other_column_defs}
          FROM {other_schema}.bond
         WHERE id
            IN (
                SELECT bond
                  FROM {other_schema}.bondset_bond
                 WHERE bondset = {other_bondset}
               )
           AND id
        NOT IN (
                SELECT bond
                  FROM {schema}.bondset_bond
                 WHERE bondset = {bondset}
               )
        """
        for row in self.db.execute(sql):
            deleted[row["id"]] = row[1:]

        if len(deleted) > 0:
            result["columns in deleted rows"] = row.keys()[1:]
            result["deleted"] = deleted

        # Detach the other database if needed
        if detach:
            self.system_db.detach(other_db)

        return result

    def get_as_dict(self, *args):
        """Return the bond data as a Python dictionary of lists.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        dict(str: [])
            A dictionary whose keys are the column names and values as lists
        """
        rows = self.bonds(*args)
        columns = [x[0] for x in rows.description]
        data = {key: [] for key in columns}
        for row in rows:
            for key, value in zip(columns, row):
                data[key].append(value)

        return data

    def get_bond(self, i, j=None, force=False):
        """Return the row for a bond.

        Parameters
        ----------
        i : int
            Either the first atom in the bond or the bond id.
        j : int = None
            If not None, then the second atom in the bond.
        force : bool = False
            If true, ignore missing bonds.

        Returns
        -------
        row : SQLite3.Row
            The Row object for the bond
        """
        if j is None:
            sql = (
                "SELECT bond.* FROM bond, bondset_bond"
                " WHERE id = ?"
                "   AND bondset_bond.bond = bond.id"
                "   AND bondset_bond.bondset = ?"
            )
            self.cursor.execute(sql, (i, self.bondset))
            row = self.cursor.fetchone()
            if row is None and not force:
                raise KeyError(f"No bond id = {i} found")
        else:
            if i > j:
                j, i = i, j
            sql = (
                "SELECT bond.* FROM bond, bondset_bond"
                " WHERE bond.i = ? AND bond.j = ?"
                "   AND bondset_bond.bondset = ?"
            )
            self.cursor.execute(sql, (i, j, self.bondset))
            row = self.cursor.fetchone()
            if row is None and not force:
                raise KeyError(f"No bond from {i} to {j} found")
        return row

    def get_column(self, key):
        """Return a column of data.

        Parameters
        ----------
        key : str
            The column (attribute) to getLogger

        Returns
        -------
        column : Column
            The column object requested.
        """
        if key == "i" or key == "j" or key == "id":
            sql = f"""
            SELECT {key}
               FROM bond, bondset_bond
              WHERE bond.id = bondset_bond.bond
                AND bondset_bond.bondset = {self.bondset}
            """
            return _FrozenColumn(self, key, sql)
        else:
            sql = f"""
            SELECT id, {key}
               FROM bond, bondset_bond
              WHERE bond.id = bondset_bond.bond
                AND bondset_bond.bondset = {self.bondset}
            """
            return _Column(self, key, sql)

    def get_column_data(self, key):
        """Return a column of data as a Python list.

        Parameters
        ----------
        key : str
            The column (attribute) to getLogger

        Returns
        -------
        column : Column
            The column object requested.
        """
        sql = f"""
        SELECT {key}
          FROM bond, bondset_bond
         WHERE bond.id = bondset_bond.bond
           AND bondset_bond.bondset = {self.bondset}
        """
        return [row[0] for row in self.db.execute(sql)]

    def get_ids(self, *args):
        """The ids of the bonds.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        [int]
            The ids of the requested bonds.
        """

        sql = (
            "SELECT bond.id FROM bond, bondset_bond"
            " WHERE bond.id = bondset_bond.bond"
            "   AND bondset_bond.bondset = ?"
        )

        parameters = [self.bondset]
        if len(args) > 0:
            for col, op, value in grouped(args, 3):
                if op == "==":
                    op = "="
                sql += f' AND "{col}" {op} ?'
                parameters.append(value)

        return [x[0] for x in self.db.execute(sql, parameters)]

    def import_bonds(self, ids):
        """Import existing bonds into this configuration.

        Parameters
        ----------
        ids : iterable(int)
            The ids of the bonds to import.
        """
        table = _Table(self.system_db, "bondset_bond")
        table.append(bondset=self.bondset, bond=ids)

    def to_dataframe(self):
        """Return the bonds as a Pandas Dataframe."""
        cursor = self.bonds()
        data = {row[0]: row[1:] for row in cursor}
        columns = [x[0] for x in cursor.description[1:]]

        df = pandas.DataFrame.from_dict(data, orient="index", columns=columns)

        return df


class _SubsetBonds(_Bonds):
    """The Bonds class describes the bonds in a subset

    :meta public:
    """

    def __init__(self, configuration, subset_id):
        self._sid = subset_id

        # Caching
        self._template_id = None
        self._template = None

        super().__init__(configuration)

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        raise NotImplementedError()

    @property
    def ids(self):
        """The ids of the bonds in the subset."""
        return self.get_bond_ids()

    @property
    def n_bonds(self):
        """The number of bonds."""
        sql = """
        SELECT COUNT(*) FROM bond
         WHERE id IN (SELECT bond FROM bondset_bond WHERE bondset = ?)
           AND i IN (SELECT atom FROM subset_atom WHERE subset = ?)
           AND j IN (SELECT atom FROM subset_atom WHERE subset = ?)
        """
        self.cursor.execute(sql, (self.bondset, self.subset_id))
        return self.cursor.fetchone()[0]

    def append(self, **kwargs):
        """Append one or more bonds.

        Not currently allowed in subsets."""
        raise NotImplementedError("Can't add bonds in a subset yet.")

    def bonds(self, *args):
        """Returns an iterator over the bonds.

        Parameters
        ----------
        args : [str]
            Added selection criteria for the SQL, one word at a time.

        Returns
        -------
        sqlite3.Cursor
            A cursor that returns sqlite3.Row objects for the bonds.
        """
        sql = """
        SELECT * FROM bond
         WHERE id IN (SELECT bond FROM bondset_bond WHERE bondset = ?)
           AND i IN (SELECT atom FROM subset_atom WHERE subset = ?)
           AND j IN (SELECT atom FROM subset_atom WHERE subset = ?)
        """
        return self.db.execute(sql, (self.bondset, self.subset_id))

    def contains_bond(self, i, j):
        """Whether there is a bond between atoms i and j."""
        # get canonical order
        if i > j:
            j, i = i, j
        sql = """
        SELECT COUNT(*) FROM bond
         WHERE i = ? AND j = ?
           AND id IN (SELECT bond FROM bondset_bond WHERE bondset = ?)
           AND i IN (SELECT atom FROM subset_atom WHERE subset = ?)
           AND j IN (SELECT atom FROM subset_atom WHERE subset = ?)
        """
        self.cursor.execute(sql, (i, j, self.bondset, self.subset_id))
        return self.cursor.fetchone()[0] == 1

    def delete_bond(self, i, j=None, force=False):
        """Delete a bond, if present.


        Not currently allowed in subsets."""
        raise NotImplementedError("Can't delete bonds from a subset yet.")

    def delete(self, atoms=None):
        """Deletes all the bonds, optionally from atoms.

        Not currently allowed in subsets."""
        raise NotImplementedError("Can't add bonds in a subset yet.")

    def diff(self, other):
        """Difference between these bonds and another

        Not currently allowed in subsets."""
        raise NotImplementedError("Can't add bonds in a subset yet.")

    def get_bond(self, i, j=None, force=False):
        """Return the row for a bond.

        Parameters
        ----------
        i : int
            Either the first atom in the bond or the bond id.
        j : int = None
            If not None, then the second atom in the bond.
        force : bool = False
            If true, ignore missing bonds.

        Returns
        -------
        row : SQLite3.Row
            The Row object for the bond
        """
        if j is None:
            sql = """
            SELECT * FROM bond
             WHERE id = ?
               AND id IN (SELECT bond FROM bondset_bond WHERE bondset = ?)
               AND i IN (SELECT atom FROM subset_atom WHERE subset = ?)
               AND j IN (SELECT atom FROM subset_atom WHERE subset = ?)
            """
            self.cursor.execute(sql, (i, self.bondset, self.subset_id, self.subset_id))
            row = self.cursor.fetchone()
            if row is None and not force:
                raise KeyError(f"No bond id = {i} found")
        else:
            if i > j:
                j, i = i, j
            sql = """
            SELECT COUNT(*) FROM bond
             WHERE i = ? AND j = ?
               AND id IN (SELECT bond FROM bondset_bond WHERE bondset = ?)
               AND i IN (SELECT atom FROM subset_atom WHERE subset = ?)
               AND j IN (SELECT atom FROM subset_atom WHERE subset = ?)
            """
            self.cursor.execute(
                sql, (i, j, self.bondset, self.subset_id, self.subset_id)
            )
            row = self.cursor.fetchone()
            if row is None and not force:
                raise KeyError(f"No bond from {i} to {j} found")
        return row

    def get_column(self, key):
        """Return a column of data.

        Parameters
        ----------
        key : str
            The column (attribute) to getLogger

        Returns
        -------
        column : Column
            The column object requested.
        """
        if key == "i" or key == "j" or key == "id":
            sql = f"""
            SELECT {key} FROM bond
               AND id IN (
                  SELECT bond FROM bondset_bond WHERE bondset = {self.bondset}
               )
               AND i IN (
                  SELECT atom FROM subset_atom WHERE subset = {self.subset_id}
               )
               AND j IN (
                  SELECT atom FROM subset_atom WHERE subset = {self.subset_id}
            )
            """
            return _FrozenColumn(self, key, sql)
        else:
            sql = f"""
            SELECT id, {key} FROM bond
               AND id IN (
                  SELECT bond FROM bondset_bond WHERE bondset = {self.bondset}
               )
               AND i IN (
                  SELECT atom FROM subset_atom WHERE subset = {self.subset_id}
               )
               AND j IN (
                  SELECT atom FROM subset_atom WHERE subset = {self.subset_id}
            )
            """
            return _Column(self, key, sql)

    def get_column_data(self, key):
        """Return a column of data as a Python list.

        Parameters
        ----------
        key : str
            The column (attribute) to getLogger

        Returns
        -------
        column : Column
            The column object requested.
        """
        sql = f"""
        SELECT {key} FROM bond
           AND id IN (
              SELECT bond FROM bondset_bond WHERE bondset = {self.bondset}
           )
           AND i IN (
              SELECT atom FROM subset_atom WHERE subset = {self.subset_id}
           )
           AND j IN (
              SELECT atom FROM subset_atom WHERE subset = {self.subset_id}
        )
        """
        return [row[0] for row in self.db.execute(sql)]



# -*- coding: utf-8 -*-

import logging
import math

import numpy

logger = logging.getLogger(__name__)


def cos(value):
    return math.cos(math.radians(value))


def sin(value):
    return math.sin(math.radians(value))


def dot(va, vb):
    sum = 0.0
    for a, b in zip(va, vb):
        sum += a * b
    return sum


class Cell(object):
    """A class to handle cell parameters and their transformations."""

    def __init__(self, a, b, c, alpha, beta, gamma):
        self._parameters = [a, b, c, alpha, beta, gamma]

    def __getitem__(self, key):
        """Allow [] to access the data!"""
        return self._parameters[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._parameters[key] = value

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._parameters)

    def __len__(self) -> int:
        """The len() command"""
        return len(self._parameters)

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self.equal(other, tol=1.0e-12)

    def __repr__(self):
        """The representation of this object"""
        return repr(self._parameters)

    def __str__(self):
        """The pretty string representation of this object"""
        return str(self._parameters)

    @property
    def a(self):
        """The length of the first cell vector."""
        return self._parameters[0]

    @a.setter
    def a(self, value):
        self._parameters[0] = value
        return list(self._parameters)

    @property
    def b(self):
        """The length of the second cell vector."""
        return self._parameters[1]

    @b.setter
    def b(self, value):
        self._parameters[1] = value
        return list(self._parameters)

    @property
    def c(self):
        """The length of the third cell vector."""
        return self._parameters[2]

    @c.setter
    def c(self, value):
        self._parameters[2] = value
        return list(self._parameters)

    @property
    def alpha(self):
        """The angle between b and c."""
        return self._parameters[3]

    @alpha.setter
    def alpha(self, value):
        self._parameters[3] = value
        return list(self._parameters)

    @property
    def beta(self):
        """The angle between a and c."""
        return self._parameters[4]

    @beta.setter
    def beta(self, value):
        self._parameters[4] = value
        return list(self._parameters)

    @property
    def gamma(self):
        """The angle between a and b."""
        return self._parameters[5]

    @gamma.setter
    def gamma(self, value):
        self._parameters[5] = value
        return list(self._parameters)

    @property
    def parameters(self):
        """The cell parameters as a list."""
        return list(self._parameters)

    @parameters.setter
    def parameters(self, value):
        if len(value) != 6:
            raise ValueError("parameters must be of length 6")
        self._parameters = list(value)

    @property
    def volume(self):
        """The volume of the cell."""
        a, b, c, alpha, beta, gamma = self.parameters
        # Roundoff errors!
        value = cos(alpha) * cos(beta) * cos(gamma)
        if value < 0.0 and abs(value) < 1.0e-8:
            value = 0.0
        return a * b * c * (
            1 - cos(alpha) ** 2 - cos(beta) ** 2 - cos(gamma) ** 2
        ) + 2 * math.sqrt(value)

    def equal(self, other, tol=1.0e-6):
        """Check if we are equal to another iterable to within a tolerance.

        Parameters
        ----------
        other : iterable
            The other object to check against
        tol : float = 1.0e-06
            The tolerance for comparing floating point numbers.

        Returns
        -------
        equals : bool
            Boolean indicating whether the two are equal.
        """
        if len(other) != 6:
            return False

        for i, j in zip(self, other):
            if abs(i - j) > tol:
                return False

        return True

    def reciprocal_lengths(self):
        """The length of the reciprocal space lattice vectors, physics definition

        Returns
        -------
        [float*3]
            The 3 vector lengths
        """
        v = self.reciprocal_vectors()
        return [
            math.sqrt(dot(v[0], v[0])),
            math.sqrt(dot(v[1], v[1])),
            math.sqrt(dot(v[2], v[2])),
        ]

    def reciprocal_vectors(self, as_array=False):
        """The reciprocal space lattice vectors. Physics definition with 2 pi

        Parameters
        ----------
        as_array : bool = False
            Whether to return a numpy array or Python lists

        Returns
        -------
        transform : [N][float*3] or ndarray
            The transformation matrix
        """
        a, b, c, alpha, beta, gamma = self.parameters

        ca = cos(alpha)
        cb = cos(beta)
        cg = cos(gamma)
        sg = sin(gamma)

        twopi = 2 * math.pi
        V = a * b * c * math.sqrt(1 - ca ** 2 - cb ** 2 - cg ** 2 + 2 * ca * cb * cg)
        # Transpose...
        # [1 / a, -cg / (a * sg), b * c * (ca * cg - cb) / (V * sg)],
        # [0, 1 / (b * sg), a * c * (cb * cg - ca) / (V * sg)],
        # [0, 0, a * b * sg / V]
        T = [
            [twopi / a, 0, 0],
            [-twopi * cg / (a * sg), twopi / (b * sg), 0],
            [
                twopi * b * c * (ca * cg - cb) / (V * sg),
                twopi * a * c * (cb * cg - ca) / (V * sg),
                twopi * a * b * sg / V,
            ],
        ]  # yapf: disable

        if as_array:
            return numpy.array(T)
        else:
            return T

    def to_cartesians(self, uvw, as_array=False):
        """Convert fraction coordinates to Cartesians

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        uvw : [N][3*float] or ndarray
            The fractional coordinates.

        Returns
        -------
        xyz : [N][float*3] or ndarray
            The Cartesian coordinates.
        """
        if isinstance(uvw, numpy.ndarray):
            UVW = uvw
        else:
            UVW = numpy.array(uvw)

        T = self.to_cartesians_transform(as_array=True)
        XYZ = UVW @ T

        if as_array:
            return XYZ
        else:
            return XYZ.tolist()

    def to_cartesians_transform(self, as_array=False):
        """Matrix to convert fractional coordinates to Cartesian.

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        as_array : bool = False
            Whether to return a numpy array or Python lists

        Returns
        -------
        transform : [N][float*3] or ndarray
            The transformation matrix
        """
        a, b, c, alpha, beta, gamma = self.parameters

        ca = cos(alpha)
        cb = cos(beta)
        cg = cos(gamma)
        sg = sin(gamma)

        V = a * b * c * math.sqrt(1 - ca ** 2 - cb ** 2 - cg ** 2 + 2 * ca * cb * cg)
        # Transpose of ...
        # [a, b * cg, c * cb],
        # [0, b * sg, c * (ca - cb * cg) / sg],
        # [0, 0, V / (a * b * sg)]
        T = [
            [a, 0, 0],
            [b * cg, b * sg, 0],
            [c * cb, c * (ca - cb * cg) / sg, V / (a * b * sg)],
        ]  # yapf: disable

        if as_array:
            return numpy.array(T)
        else:
            return T

    def to_fractionals(self, xyz, as_array=False):
        """Convert Cartesian coordinates to fractional.

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        xyz : [N][3*float] or ndarray
            The Cartesian coordinates.

        Returns
        -------
        uvw : [N][float*3] or ndarray
            The ractional coordinates.
        """
        if isinstance(xyz, numpy.ndarray):
            XYZ = xyz
        else:
            XYZ = numpy.array(xyz)

        T = self.to_fractionals_transform(as_array=True)
        UVW = XYZ @ T

        if as_array:
            return UVW
        else:
            return UVW.tolist()

    def to_fractionals_transform(self, as_array=False):
        """Matrix to convert Cartesian coordinates to fractional.

        see https://en.wikipedia.org/wiki/Fractional_coordinates for a
        description.

        Parameters
        ----------
        as_array : bool = False
            Whether to return a numpy array or Python lists

        Returns
        -------
        transform : [N][float*3] or ndarray
            The transformation matrix
        """
        a, b, c, alpha, beta, gamma = self.parameters

        ca = cos(alpha)
        cb = cos(beta)
        cg = cos(gamma)
        sg = sin(gamma)

        V = a * b * c * math.sqrt(1 - ca ** 2 - cb ** 2 - cg ** 2 + 2 * ca * cb * cg)
        # Transpose...
        # [1 / a, -cg / (a * sg), b * c * (ca * cg - cb) / (V * sg)],
        # [0, 1 / (b * sg), a * c * (cb * cg - ca) / (V * sg)],
        # [0, 0, a * b * sg / V]
        T = [
            [1 / a, 0, 0],
            [-cg / (a * sg), 1 / (b * sg), 0],
            [
                b * c * (ca * cg - cb) / (V * sg),
                a * c * (cb * cg - ca) / (V * sg),
                a * b * sg / V,
            ],
        ]  # yapf: disable

        if as_array:
            return numpy.array(T)
        else:
            return T

    def vectors(self, as_array=False):
        """The cell or lattice vectors.

        Parameters
        ----------
        as_array : bool = False
            Whether to return a numpy array or Python lists

        Returns
        -------
        transform : [N][float*3] or ndarray
            The transformation matrix
        """
        return self.to_cartesians_transform(as_array=as_array)

    def from_vectors(self, vectors):
        """Set the cell parameters from the lattice vectors.

        Parameters
        ----------
        vectors : [[float*3]*3]
            The lattice vectors as a list [a, b, c]
        """
        va = vectors[0]
        vb = vectors[1]
        vc = vectors[2]
        a = math.hypot(*va)
        b = math.hypot(*vb)
        c = math.hypot(*vc)
        alpha = math.degrees(math.acos(dot(vb, vc) / (b * c)))
        beta = math.degrees(math.acos(dot(va, vc) / (a * c)))
        gamma = math.degrees(math.acos(dot(va, vb) / (a * b)))

        self.parameters = [a, b, c, alpha, beta, gamma]


class _Cell(Cell):
    """A class for handling cell parameters as part of MolSystem.

    Provides all the functionality of the Cell class, but keeps the cell
    data in the database.

    :meta public:
    """

    def __init__(self, configuration):
        """Initialize from the database.

        Parameters
        ----------
        system_db : SystemDB
            The SystemDB instance that we are working with.
        _id : int
            The id of this particular cell.
        """
        self._configuration = configuration
        self._system = self._configuration.system
        self._system_db = self._system.system_db
        self._id = configuration.cell_id

        self.cursor.execute(
            "SELECT a, b, c, alpha, beta, gamma FROM cell WHERE id = ?", (self._id,)
        )
        super().__init__(*self.cursor.fetchone())

    def __enter__(self):
        """Copy the tables to a backup for a 'with' statement."""
        self.system_db["cell"].__enter__()
        return self

    def __exit__(self, etype, value, traceback):
        """Handle returning from a 'with' statement."""
        if etype is None:
            self.configuration.version = self.configuration.version + 1
        return self.system_db["cell"].__exit__(etype, value, traceback)

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        # This gets rid if LGTM warning...
        return self.equal(other, tol=1.0e-12)

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._parameters[key] = value
        self._save()

    @property
    def cursor(self):
        return self.system_db.cursor

    @property
    def db(self):
        return self.system_db.db

    @property
    def configuration(self):
        """Return the configuration."""
        return self._configuration

    @property
    def id(self):
        """The id of this cell."""
        return self._id

    @property
    def a(self):
        """The length of the first cell vector."""
        return self._parameters[0]

    @a.setter
    def a(self, value):
        self._parameters[0] = value
        self._save()
        return list(self._parameters)

    @property
    def b(self):
        """The length of the second cell vector."""
        return self._parameters[1]

    @b.setter
    def b(self, value):
        self._parameters[1] = value
        self._save()
        return list(self._parameters)

    @property
    def c(self):
        """The length of the third cell vector."""
        return self._parameters[2]

    @c.setter
    def c(self, value):
        self._parameters[2] = value
        self._save()
        return list(self._parameters)

    @property
    def alpha(self):
        """The angle between b and c."""
        return self._parameters[3]

    @alpha.setter
    def alpha(self, value):
        self._parameters[3] = value
        self._save()
        return list(self._parameters)

    @property
    def beta(self):
        """The angle between a and c."""
        return self._parameters[4]

    @beta.setter
    def beta(self, value):
        self._parameters[4] = value
        self._save()
        return list(self._parameters)

    @property
    def gamma(self):
        """The angle between a and b."""
        return self._parameters[5]

    @gamma.setter
    def gamma(self, value):
        self._parameters[5] = value
        self._save()
        return list(self._parameters)

    @property
    def parameters(self):
        """The cell parameters as a list."""
        return list(self._parameters)

    @parameters.setter
    def parameters(self, value):
        if len(value) != 6:
            raise ValueError("parameters must be of length 6")
        self._parameters = list(value)
        self._save()
        return list(self._parameters)

    @property
    def system(self):
        """Return the System object that contains this cell."""
        return self._system

    @property
    def system_db(self):
        """Return the SystemDB object that contains this cell."""
        return self._system_db

    def diff(self, other):
        """Difference between these cell and another

        Parameters
        ----------
        other : _Cell
            The other cell to diff against

        Result
        ------
        result : Dict
            The differences, described in a dictionary
        """
        result = {}

        # Check the columns
        columns = self._columns()
        other_columns = other._columns()

        column_defs = ", ".join(columns)
        other_column_defs = ", ".join(other_columns)

        if columns == other_columns:
            column_def = column_defs
        else:
            added = columns - other_columns
            if len(added) > 0:
                result["columns added"] = list(added)
            deleted = other_columns - columns
            if len(deleted) > 0:
                result["columns deleted"] = list(deleted)

            in_common = other_columns & columns
            if len(in_common) > 0:
                column_def = ", ".join(in_common)
            else:
                # No columns shared
                return result

        # Need to check the contents of the tables. See if they are in the same
        # database or if we need to attach the other database temporarily.
        db = self.system_db
        other_db = other.system_db

        detach = False
        schema = self.schema
        if db.filename != other_db.filename:
            if db.is_attached(other_db):
                other_schema = db.attached_as(other_db)
            else:
                # Attach the other system_db in order to do comparisons.
                other_schema = self.system_db.attach(other_db)
                detach = True
        else:
            other_schema = other.schema

        _id = self.id
        other_id = other.id

        changed = {}
        last = None
        sql = f"""
        SELECT * FROM
        (
          SELECT {column_def}
            FROM {other_schema}.bond
           WHERE id = {other_id}
          EXCEPT
          SELECT {column_def}
            FROM {schema}.bond
           WHERE id = {_id}
        )
         UNION ALL
        SELECT * FROM
        (
          SELECT {column_def}
            FROM {schema}.bond
           WHERE id = {_id}
          EXCEPT
          SELECT {column_def}
            FROM {other_schema}.bond
           WHERE id = {other_id}
        )
        ORDER BY id
        """

        for row in self.db.execute(sql):
            if last is None:
                last = row
            elif row["id"] == last["id"]:
                # changes = []
                changes = set()
                for k1, v1, v2 in zip(last.keys(), last, row):
                    if v1 != v2:
                        changes.add((k1, v1, v2))
                changed[row["id"]] = changes
                last = None
            else:
                last = row
        if len(changed) > 0:
            result["changed"] = changed

        # See about the rows added
        added = {}
        sql = f"""
        SELECT {column_defs}
          FROM {schema}.bond
         WHERE id = {_id}
           AND id <> {other_id}
        """
        for row in self.db.execute(sql):
            added[row["id"]] = row[1:]

        if len(added) > 0:
            result["columns in added rows"] = row.keys()[1:]
            result["added"] = added

        # See about the rows deleted
        deleted = {}
        sql = f"""
        SELECT {other_column_defs}
          FROM {other_schema}.bond
         WHERE id = {other_id}
           AND id <> {_id}
        """
        for row in self.db.execute(sql):
            deleted[row["id"]] = row[1:]

        if len(deleted) > 0:
            result["columns in deleted rows"] = row.keys()[1:]
            result["deleted"] = deleted

        # Detach the other database if needed
        if detach:
            self.system_db.detach(other_db)

        return result

    def _save(self):
        """Save our values in the database."""
        parameters = list(self._parameters)
        parameters.append(self.id)
        self.cursor.execute(
            "UPDATE cell SET a=?, b=?, c=?, alpha=?, beta=?, gamma=?" "WHERE id = ?",
            parameters,
        )
        self.db.commit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401

"""Tests for handling Molfiles."""

text1 = """\
L-Alanine
GSMACCS-II07189510252D 1 0.00366 0.00000 0
Figure 1, J. Chem. Inf. Comput. Sci., Vol 32, No. 3., 1992
0 0 0 0 0 999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 6 5 0 0 1
M  V30 BEGIN ATOM
M  V30 1 C -0.6622 0.5342 0 0 CFG=2
M  V30 2 C 0.6622 -0.3 0 0
M  V30 3 C -0.7207 2.0817 0 0 MASS=13
M  V30 4 N -1.8622 -0.3695 0 0 CHG=1
M  V30 5 O 0.622 -1.8037 0 0
M  V30 6 O 1.9464 0.4244 0 0 CHG=-1
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 1 2
M  V30 2 1 1 3 CFG=1
M  V30 3 1 1 4
M  V30 4 2 2 5
M  V30 5 1 2 6
M  V30 END BOND
M  V30 END CTAB
M  END
"""

atoms1 = """\
   atno  formal_charge  configuration       x       y    z
1     6              0              1 -0.6622  0.5342  0.0
2     6              0              1  0.6622 -0.3000  0.0
3     6              0              1 -0.7207  2.0817  0.0
4     7              1              1 -1.8622 -0.3695  0.0
5     8              0              1  0.6220 -1.8037  0.0
6     8             -1              1  1.9464  0.4244  0.0"""

bonds1 = """\
   i  j  bondorder
1  1  2          1
2  1  3          1
3  1  4          1
4  2  5          2
5  2  6          1"""

text2 = """\
L-Alanine
PSSEAMM_WF09112015563D
Exported from SEAMM
  0  0  0     0  0            999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 6 5 0 0 0
M  V30 BEGIN ATOM
M  V30 1 C -0.6622 0.5342 0.0 0
M  V30 2 C 0.6622 -0.3 0.0 0
M  V30 3 C -0.7207 2.0817 0.0 0
M  V30 4 N -1.8622 -0.3695 0.0 0
M  V30 5 O 0.622 -1.8037 0.0 0
M  V30 6 O 1.9464 0.4244 0.0 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 1 2
M  V30 2 1 1 3
M  V30 3 1 1 4
M  V30 4 2 2 5
M  V30 5 1 2 6
M  V30 END BOND
M  V30 END CTAB
M  END
"""

text3 = """\
acetic acid
PSSEAMM_WF09112018063D
Exported from SEAMM
  0  0  0     0  0            999 V3000
M  V30 BEGIN CTAB
M  V30 COUNTS 8 7 0 0 0
M  V30 BEGIN ATOM
M  V30 1 C 1.0797 0.0181 -0.0184 0
M  V30 2 H 0.5782 3.1376 0.2813 0
M  V30 3 H 0.7209 -0.6736 -0.7859 0
M  V30 4 H 0.7052 -0.3143 0.9529 0
M  V30 5 C 0.5713 1.3899 -0.3161 0
M  V30 6 O -0.1323 1.7142 -1.2568 0
M  V30 7 O 0.9757 2.297 0.5919 0
M  V30 8 H 2.1724 0.0161 -0.0306 0
M  V30 END ATOM
M  V30 BEGIN BOND
M  V30 1 1 1 2
M  V30 2 1 1 3
M  V30 3 1 1 4
M  V30 4 1 1 5
M  V30 5 2 5 6
M  V30 6 1 5 7
M  V30 7 1 7 8
M  V30 END BOND
M  V30 END CTAB
M  END
"""


def test_from_text(configuration):
    """Create a configuration from the text of a Molfile"""
    configuration.from_molfile_text(text1)
    if str(configuration.atoms) != atoms1:
        print(configuration.atoms)
    if str(configuration.bonds) != bonds1:
        print(configuration.bonds)
    assert str(configuration.atoms) == atoms1
    assert str(configuration.bonds) == bonds1


def test_to_text(configuration):
    """Create the text of a Molfile from a configuration"""
    configuration.from_molfile_text(text1)
    # Second line contains the current date/time
    text_sv = configuration.to_molfile_text()
    text = text_sv.splitlines()
    del text[1]
    text = "\n".join(text)

    tmp = text2.splitlines()
    del tmp[1]
    tmp = "\n".join(tmp)

    if text != tmp:
        print(text_sv)
    assert text == tmp


def test_to_text2(AceticAcid):
    """Write a manually created configuration to molfile"""
    # Second line contains the current date/time
    text_sv = AceticAcid.to_molfile_text()
    text = text_sv.splitlines()
    del text[1]
    text = "\n".join(text)

    tmp = text3.splitlines()
    del tmp[1]
    tmp = "\n".join(tmp)

    if text != tmp:
        print(text_sv)
    assert text == tmp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for properties."""

import pytest  # noqa: F401


def test_empty(empty_db):
    """Test empty properties db."""
    assert empty_db.properties.known_properties() == []


def test_add(empty_db):
    """Test adding properties db."""
    properties = empty_db.properties
    assert not properties.exists("dipole moment")
    properties.add("dipole moment")
    assert properties.known_properties() == ["dipole moment"]
    assert properties.exists("dipole moment")


def test_put(AceticAcid):
    """Test putting some properties and values in."""
    cid = AceticAcid.id
    properties = AceticAcid.system_db.properties
    properties.add("dipole moment")
    properties.put(cid, "dipole moment", 2.0)
    result = properties.get(cid, "dipole moment")
    assert result == 2.0


def test_query(AceticAcid):
    """Test a simple query"""
    cid = AceticAcid.id
    properties = AceticAcid.system_db.properties
    properties.add("dipole moment")
    properties.put(cid, "dipole moment", 2.0)

    assert properties.query(what=["dipole moment"]) == {"dipole moment": [2.0]}


def test_query2(properties):
    """Test a simple query"""
    result = properties.query(
        "float_0",
        "between",
        -10.0,
        -9.9,
        what=["float_8", "str_9"],
    )
    assert result["str_9"] == [
        "string 400",
        "string 401",
        "string 402",
        "string 403",
        "string 404",
        "string 405",
        "string 406",
        "string 407",
        "string 408",
        "string 409",
        "string 410",
    ]
    assert result["float_8"] == [
        -2.0,
        -1.99,
        -1.98,
        -1.97,
        -1.96,
        -1.95,
        -1.94,
        -1.93,
        -1.92,
        -1.91,
        -1.9,
    ]



# -*- coding: utf-8 -*-
import logging

from .atoms import _SubsetAtoms
from .bonds import _SubsetBonds
from .openbabel import OpenBabelMixin
from .smiles import SMILESMixin
from .template import _Template

logger = logging.getLogger(__name__)


class _Subset(SMILESMixin, OpenBabelMixin, object):
    """:meta public:
    A class providing the API for a subset.

    Parameters
    ----------
    sid : int
        The id of the subset in the subset table.
    logger : logging.Logger = logger
        A logger to use in place of the one from this module.
    """

    def __init__(self, system_db, sid, logger=logger):
        self._system_db = system_db
        self._id = sid
        self._logger = logger

        self._template = None
        self._configuration = None

    @property
    def atoms(self):
        """The atoms for this subset.

        Returns
        -------
        _SubsetAtoms
            The atoms for the subset.
        """
        return _SubsetAtoms(self.configuration, self.id)

    @property
    def bonds(self):
        """The bonds for this subset.

        Returns
        -------
        _SubsetBonds
            The bonds for the subset.
        """
        return _SubsetBonds(self.configuration, self.id)

    @property
    def category(self):
        """The category of this subset."""
        return self.template.category

    @property
    def cursor(self):
        """The a cursor for the database."""
        return self.system_db.cursor

    @property
    def configuration(self):
        """The configuration for this subset."""
        if self._configuration is None:
            sql = "SELECT configuration FROM subset WHERE id = ?"
            self.cursor.execute(sql, (self._id,))
            cid = self.cursor.fetchone()[0]
            self._configuration = self.system_db.get_configuration(cid)
        return self._configuration

    @property
    def db(self):
        """The database connection."""
        return self.system_db.db

    @property
    def id(self):
        """Return the id of the subset."""
        return self._id

    @property
    def system_db(self):
        """The system_db that we belong to."""
        return self._system_db

    @property
    def template(self):
        """The template for this subset."""
        if self._template is None:
            sql = "SELECT template FROM subset WHERE id = ?"
            self.cursor.execute(sql, (self._id,))
            tid = self.cursor.fetchone()[0]
            self._template = _Template(self.system_db, tid)
        return self._template



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the SystemDB class.

These tests are standalone, and test that the class heirarchy works to
a minimal extent, and that we can make a system with atoms and bonds.

Once these test work then conftest.py should function and the other
tests can explore the functionlaity in more detail.
"""

import math
import pytest  # noqa: F401

from molsystem import SystemDB


def test_construction():
    """Simplest test that we can make a SystemDB object"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    assert str(type(db)) == "<class 'molsystem.system_db.SystemDB'>"
    assert db.n_systems == 0
    db.close()


def test_system():
    """Test that we can get a System object"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    system = db.create_system(name="default")
    assert str(type(system)) == "<class 'molsystem.system._System'>"
    assert system.n_configurations == 0
    db.close()


def test_configuration():
    """Test that we can get a Configuration object"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    system = db.create_system(name="default")
    configuration = system.create_configuration()
    _type = str(type(configuration))
    assert _type == "<class 'molsystem.configuration._Configuration'>"
    db.close()


def test_atoms():
    """Test that we can get an Atoms object"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    system = db.create_system(name="default")
    configuration = system.create_configuration()
    atoms = configuration.atoms
    assert str(type(atoms)) == "<class 'molsystem.atoms._Atoms'>"
    db.close()


def test_adding_atoms():
    """Test that we can add atoms"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    system = db.create_system(name="default")
    configuration = system.create_configuration()
    atoms = configuration.atoms
    symbol = ["Ar"] * 4
    X = [2.0, 0.0, -2.0, 0.0]
    Y = [0.0, 2.0, 0.0, -2.0]
    Z = [0.0] * 4
    ids = atoms.append(symbol=symbol, x=X, y=Y, z=Z)
    assert atoms.n_atoms == 4
    assert ids == [1, 2, 3, 4]
    db.close()


def test_bonds():
    """Test that we can get an Bonds object"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    system = db.create_system(name="default")
    configuration = system.create_configuration()
    bonds = configuration.bonds
    assert str(type(bonds)) == "<class 'molsystem.bonds._Bonds'>"
    db.close()


def test_adding_bonds():
    """Test that we can add bonds"""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    system = db.create_system(name="default")
    configuration = system.create_configuration()

    # TIP3P
    r0 = 0.9572
    theta0 = 104.52

    # H locations are ±x, 0, z
    x = r0 * math.sin(math.radians(theta0 / 2))
    z = r0 * math.cos(math.radians(theta0 / 2))

    X = [0.0, x, -x]
    Y = [0.0, 0.0, 0.0]
    Z = [0.0, z, z]

    atno = [8, 1, 1]
    name = ["O", "H1", "H2"]
    i_atom = [0, 0]
    j_atom = [1, 2]

    atoms = configuration.atoms
    atoms.add_attribute("name", "str", default="")
    atom_ids = atoms.append(atno=atno, x=X, y=Y, z=Z, name=name)

    i = [atom_ids[x] for x in i_atom]
    j = [atom_ids[x] for x in j_atom]

    bonds = configuration.bonds
    bond_ids = bonds.append(i=j, j=i)  # flipped on purpose so code orders.

    assert bonds.n_bonds == 2
    assert bond_ids == [1, 2]

    db.close()


def test_adding_reading_cif_file(amino_acids):
    """Test that we can read a cif file with many systems."""
    answer = [
        "ALA",
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLU",
        "GLN",
        "GLY",
        "HIS",
        "ILE",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "PRO",
        "SER",
        "THR",
        "TRP",
        "TYR",
        "VAL",
    ]
    assert amino_acids.n_systems == 20
    names = amino_acids.names
    if names != answer:
        print(names)
    assert names == answer


def test_accessing_systems(amino_acids):
    """Test that we can access each system."""
    answer = [
        13,
        27,
        17,
        16,
        14,
        19,
        20,
        10,
        21,
        22,
        22,
        25,
        20,
        23,
        17,
        14,
        17,
        27,
        24,
        19,
    ]
    assert amino_acids.n_systems == 20
    n_atoms = [x.configuration.n_atoms for x in amino_acids.systems]
    if n_atoms != answer:
        print(n_atoms)
    assert n_atoms == answer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the system class."""

import pytest  # noqa: F401


def test_construction(system):
    """Simplest test that we can make a System object"""
    assert str(type(system)) == "<class 'molsystem.system._System'>"
    del system


def test_create_table(system):
    """Test that we can create a table in the system."""
    table = system.create_table("table1")
    # Create a column so the table exists
    table.add_attribute("data")
    assert "table1" in system


def test_create_table_element(system):
    """Test that we can create a table using [] syntax."""
    table = system["table1"]
    # Create a column so the table exists
    table.add_attribute("data")
    assert "table1" in system


def test_delete_table(system_with_two_tables):
    """Test that we can delete a table."""
    system = system_with_two_tables
    del system["table1"]
    assert "table2" in system
    assert "table1" not in system


def test_elements(system):
    """Test that we can access the element table"""
    table = system["element"]
    assert table.n_rows == 118


def test_create_new_system(AceticAcid):
    """Test creating a new, empty system."""
    system_db = AceticAcid.system_db
    assert system_db.n_systems == 1
    system_db.create_system(name="new system")
    assert system_db.n_systems == 2
    assert system_db.system_ids == [1, 2]


def test_create_new_system_vanadium(AceticAcid):
    """Test creating a new system, filling it for vanadium."""
    system_db = AceticAcid.system_db
    assert system_db.n_systems == 1

    system = system_db.create_system(name="BCC Vanadium")
    system.create_configuration(name="default")
    assert system_db.n_systems == 2
    assert system_db.system_ids == [1, 2]

    system_db.current_system_id = 2

    configuration = system_db.system.configuration

    configuration.periodicity = 3
    configuration.coordinate_system = "fractional"
    configuration.cell.parameters = (3.03, 3.03, 3.03, 90, 90, 90)
    configuration.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol="V")

    formula, empirical_formula, Z = configuration.formula
    assert "".join(formula) == "V2"
    assert "".join(empirical_formula) == "V"
    assert Z == 2



# -*- coding: utf-8 -*-

"""Interface to openbabel."""

import logging

try:
    from openbabel import openbabel
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise

logger = logging.getLogger(__name__)


class OpenBabelMixin:
    """A mixin for handling OpenBabel via its Python interface."""

    def to_OBMol(self):
        """Return an OBMol object for the configuration, template, or subset."""
        ob_mol = openbabel.OBMol()
        for atno, xyz in zip(self.atoms.atomic_numbers, self.atoms.coordinates):
            ob_atom = ob_mol.NewAtom()
            ob_atom.SetAtomicNum(atno)
            ob_atom.SetVector(*xyz)

        # 1-based indices in ob.
        index = {j: i for i, j in enumerate(self.atoms.ids, start=1)}
        for row in self.bonds.bonds():
            ob_mol.AddBond(index[row["i"]], index[row["j"]], row["bondorder"])

        if self.__class__.__name__ == " _Configuration":
            ob_mol.SetTotalCharge(self.charge)
            if self.spin_multiplicity is None:
                n_electrons = sum(self.atoms.atomic_numbers) - self.charge
                if n_electrons % 2 == 0:
                    multiplicity = 1
                else:
                    multiplicity = 2
                self.spin_multiplicity = multiplicity
                ob_mol.SetTotalSpinMultiplicity(self.spin_multiplicity)

        return ob_mol

    def from_OBMol(self, ob_mol):
        """Transform an Open Babel molecule into the current object."""
        atnos = []
        Xs = []
        Ys = []
        Zs = []
        for ob_atom in openbabel.OBMolAtomIter(ob_mol):
            atno = ob_atom.GetAtomicNum()
            atnos.append(atno)
            Xs.append(ob_atom.x())
            Ys.append(ob_atom.y())
            Zs.append(ob_atom.z())
            logger.debug(f"atom {atno} {ob_atom.x()} {ob_atom.z()} {ob_atom.z()}")

        Is = []
        Js = []
        BondOrders = []
        for ob_bond in openbabel.OBMolBondIter(ob_mol):
            ob_i = ob_bond.GetBeginAtom()
            ob_j = ob_bond.GetEndAtom()
            i = ob_i.GetIdx()
            j = ob_j.GetIdx()
            bondorder = ob_bond.GetBondOrder()
            Is.append(i)
            Js.append(j)
            BondOrders.append(bondorder)
            logger.debug(f"bond {i} - {j} {bondorder}")

        self.clear()

        if self.__class__.__name__ == "_Configuration":
            self.charge = ob_mol.GetTotalCharge()
            self.spin_multiplicity = ob_mol.GetTotalSpinMultiplicity()

        ids = self.atoms.append(x=Xs, y=Ys, z=Zs, atno=atnos)
        i = [ids[x - 1] for x in Is]
        j = [ids[x - 1] for x in Js]
        self.bonds.append(i=i, j=j, bondorder=BondOrders)

        return self

    def find_substructures(self, template):
        """Find the substructures matching the template.

        Parameters
        ----------
        template : str, _Configuration, _Template, or _Subset
            The template, which may be a SMARTS string, or a molecular object.

        Returns
        -------
        [[int]]
            Lists of atom ids for matches.
        """
        if isinstance(template, str):
            smarts = template
        else:
            smarts = self.smiles

        ob_mol = self.to_OBMol()

        pattern = openbabel.OBSmartsPattern()
        pattern.Init(smarts)
        pattern.Match(ob_mol)
        maplist = pattern.GetUMapList()

        return [x for x in maplist]



# -*- coding: utf-8 -*-

"""A class providing a convenient interface for subsets
"""

from itertools import zip_longest
import logging

from .subset import _Subset
from .table import _Table

logger = logging.getLogger(__name__)


def grouped(iterable, n):
    "s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,...s3n-1), ..."
    return zip_longest(*[iter(iterable)] * n)


class _Subsets(_Table):
    """The Subsets class handles the subsets for a single configuration.

    Parameters
    ----------
    configuration : _Configuration
        The configuration to work with.
    logger : Logger
        The logger to use, defaults to the one for this module.

    :meta public:
    """

    def __init__(self, configuration, logger=logger):
        self._configuration = configuration
        self.logger = logger

        self._cid = configuration.id
        self._system_db = configuration.system_db

        super().__init__(self._system_db, "subset")

        self._sa_table = self._system_db["subset_atom"]

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        raise NotImplementedError()

    @property
    def n_subsets(self):
        """The number of subsets for a configuration.

        Returns
        -------
        int
            The number of subsets in the configuration.
        """
        sql = """\
        SELECT COUNT(*)
          FROM subset
         WHERE configuration = ?
        """

        self.cursor.execute(sql, (self._cid,))
        result = self.cursor.fetchone()[0]
        return result

    def append(self, n=None, **kwargs):
        """Append one or more rows

        The keywords are the names of attributes and the value to use.
        The default value for any attributes not given is used unless it is
        'None' in which case an error is thrown. It is an error if there is not
        an exisiting attribute corresponding to any given as arguments.

        Parameters
        ----------
        n : int = None
            The number of rows to create, defaults to the number of items
            in the longest attribute given.
        kwargs :
            any number <attribute name> = <value> keyword arguments giving
            existing attributes and values.

        Returns
        -------
        [int]
            The ids of the created rows.
        """
        kwargs["configuration"] = self._cid
        ids = super().append(n, **kwargs)

        return ids

    def create(self, template, atoms=None, templateatoms=None):
        """Create a subset given a template and optionally atoms.

        Parameters
        ----------
        template : int or _Template
            The template for the subset
        atoms : [int] = None
            Optional list of atom ids to connect to the subset.
        templateatoms : [int] = None
            Optional list of template atoms to connect to the atoms
            in the subset.

        Returns
        -------
        _Subset
            The subset.
        """
        if isinstance(template, int):
            tid = template
        else:
            tid = template.id
        sid = self.append(template=tid)[0]

        if atoms is not None:
            if templateatoms is not None:
                self._sa_table.append(
                    subset=sid, atom=atoms, templateatom=templateatoms
                )
            else:
                self._sa_table.append(subset=sid, atom=atoms)

        return _Subset(self.system_db, sid)

    def delete(self, ids):
        """Remove one or more subsets

        Parameters
        ----------
        ids : [int] of 'all'
            The subsets to delete.

        Returns
        -------
        None
        """
        if ids == "all":
            sql = """\
            DELETE FROM subset
             WHERE configuration = ?
            """
            self.db.execute(sql, (self._cid,))
        else:
            if isinstance(ids, int):
                self.db.execute("DELETE FROM subset WHERE id = ?", (ids,))
            else:
                self.db.executemany("DELETE FROM subset WHERE id = ?", ids)

    def generate(self, template):
        """Generate the subsets matching a full template.

        Parameters
        ----------
        template : int or _Template
           The template.

        Returns
        -------
        [_Subset]
            The subsets.
        """
        pass

    def get(self, template):
        """Get the subsets given a template.

        Parameters
        ----------
        template : int or _Template
           The template.

        Returns
        -------
        [_Subset]
            The subsets.
        """
        if isinstance(template, int):
            tid = template
        else:
            tid = template.id

        return [_Subset(self.system_db, x) for x in self.get_ids(tid)]

    def get_ids(self, template):
        """Find ids of subsets given a template.

        Parameters
        ----------
        template : int or _Template
            The template for the subset

        Returns
        -------
        [int]
            The ids of the subsets, and empty list if there are none.
        """
        if isinstance(template, int):
            tid = template
        else:
            tid = template.id

        sql = """
        SELECT id FROM subset
         WHERE template = ?
           AND configuration = ?
        """
        return [x[0] for x in self.db.execute(sql, (tid, self._cid))]

    def get_counts(self, configuration="all"):
        """Get the counts of subsets per template for configurations.

        Parameters
        ----------
        configuration : int or MolSystem._Configuration or "all"
            The configuration, as an object or its id, or "all" for all
            configurations. Default is "all".

        Returns
        -------
        [[configuration_id, name, count]]
        """

        if configuration == "all":
            sql = (
                "SELECT subset.configuration, name, count(1) FROM subset, template "
                " WHERE template = template.id "
                " GROUP BY subset.configuration, name "
                " ORDER BY subset.configuration"
            )
            return [[*x] for x in self.db.execute(sql)]
        else:
            raise NotImplementedError("get_counts only take 'all' configurations")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `bonds` in the `molsystem` package."""

import pytest  # noqa: F401


def test_construction(AceticAcid):
    """Simplest test that we can make an Bonds object"""
    configuration = AceticAcid
    bonds = configuration.bonds
    assert str(type(bonds)) == "<class 'molsystem.bonds._Bonds'>"


def test_iteration(AceticAcid):
    """Test that we can iterate over the bonds."""
    configuration = AceticAcid
    bonds = configuration.bonds
    i = []
    j = []
    order = []
    for bond in bonds.bonds():
        i.append(bond["i"])
        j.append(bond["j"])
        order.append(bond["bondorder"])

    assert order == [1, 1, 1, 1, 2, 1, 1]
    assert i == [1, 1, 1, 1, 5, 5, 7]
    assert j == [2, 3, 4, 5, 6, 7, 8]


def test_contains(AceticAcid):
    """Test that it contains a bond."""
    configuration = AceticAcid
    bonds = configuration.bonds
    assert 5, 7 in bonds
    assert 7, 5 in bonds


def test_does_not_contain(AceticAcid):
    """Test that does not contain a bond."""
    configuration = AceticAcid
    bonds = configuration.bonds
    assert 5, 8 not in bonds
    assert 8, 5 not in bonds


def test_get_item(AceticAcid):
    """Test that we can access a bond."""
    configuration = AceticAcid
    bonds = configuration.bonds
    bond = bonds.get_bond(5, 6)
    assert [*bond] == [5, 5, 6, 2]
    bond = bonds.get_bond(6, 5)
    assert [*bond] == [5, 5, 6, 2]


def test_delete_bond(AceticAcid):
    """Test that we can remove a bond."""
    configuration = AceticAcid
    bonds = configuration.bonds

    bonds.delete_bond(5, 7)
    assert bonds.n_bonds == 6


def test_add_bond(AceticAcid):
    """Test that we can remove a bond."""
    configuration = AceticAcid
    bonds = configuration.bonds
    bond = bonds.get_bond(5, 7)
    bonds.delete_bond(5, 7)
    assert bonds.n_bonds == 6
    bonds.append(bonds=bond)
    assert bonds.n_bonds == 7


def test_str(AceticAcid):
    """Test that we can get a string representation."""
    answer = """\
   i  j  bondorder
1  1  2          1
2  1  3          1
3  1  4          1
4  1  5          1
5  5  6          2
6  5  7          1
7  7  8          1"""
    configuration = AceticAcid
    bonds = configuration.bonds
    if str(bonds) != answer:
        print(str(bonds))
    assert str(bonds) == answer


def test_repr(AceticAcid):
    """Test that we can get a representation."""
    answer = """\
   i  j  bondorder
1  1  2          1
2  1  3          1
3  1  4          1
4  1  5          1
5  5  6          2
6  5  7          1
7  7  8          1"""
    configuration = AceticAcid
    bonds = configuration.bonds
    if repr(bonds) != answer:
        print(repr(bonds))
    assert repr(bonds) == answer


def test_adding_attribute(AceticAcid):
    """Test that we can add an attribute."""
    configuration = AceticAcid
    bonds = configuration.bonds
    bonds.add_attribute("name", coltype="str")
    bond = bonds.get_bond(5, 7)
    assert len(bond) == 5
    assert bond.keys() == ["id", "i", "j", "bondorder", "name"]


def test_adding_attribute_with_values(AceticAcid):
    """Test that we can add an attribute."""
    names = ["C-H", "C-H", "C-H", "C-C", "C=O", "C-O", "O-H"]
    configuration = AceticAcid
    bonds = configuration.bonds
    bonds.add_attribute("name", coltype="str", values=names)
    bond = bonds.get_bond(5, 7)
    assert len(bond) == 5
    assert bond.keys() == ["id", "i", "j", "bondorder", "name"]
    assert bond["name"] == "C-O"


def test_column(AceticAcid):
    """Test getting columns of the bond data."""
    answer = [1, 1, 1, 1, 2, 1, 1]
    configuration = AceticAcid
    bonds = configuration.bonds
    bondorders = bonds.get_column("bondorder")
    if bondorders != answer:
        print(bondorders)
    assert bondorders == answer


def test_set_column(AceticAcid):
    """Test setting columns of the bond data."""
    answer = [1, 1, 1, 1, 2, 3, 1]
    answer2 = """\
   i  j  bondorder
1  1  2          1
2  1  3          1
3  1  4          1
4  1  5          1
5  5  6          2
6  5  7          3
7  7  8          1"""
    configuration = AceticAcid
    bonds = configuration.bonds
    bondorders = bonds.get_column("bondorder")
    bondorders[5] = 3
    if bondorders != answer:
        print(bondorders)
    assert bondorders == answer
    if str(bonds) != answer2:
        print(str(bonds))
    assert str(bonds) == answer2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Fixtures for testing the 'molsystem' package."""
import math
from pathlib import Path

import pytest

from molsystem import SystemDB

path = Path(__file__).resolve().parent
data_path = path / "data"


def pytest_addoption(parser):
    parser.addoption(
        "--run-timing", action="store_true", default=False, help="run timing tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "timing: mark test as timing to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run-timing"):
        # --run-timing given in cli: do not skip timing tests
        return
    skip_timing = pytest.mark.skip(reason="need --run-timing option to run")
    for item in items:
        if "timing" in item.keywords:
            item.add_marker(skip_timing)


def mk_table(db, name="table1"):
    """Create a table with some attributes."""
    table = db.create_table(name)
    table.add_attribute("atno", coltype="int", default=-1)
    for column in ("x", "y", "z"):
        table.add_attribute(column, coltype="float")
    return table


@pytest.fixture()
def empty_db():
    """Create a system db with no systems."""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")

    yield db

    db.close()
    try:
        del db
    except:  # noqa: E722
        print("Caught error deleting the database")


@pytest.fixture()
def db(empty_db):
    """Create an empty system db."""
    system = empty_db.create_system(name="default")
    system.create_configuration(name="default")

    return empty_db


@pytest.fixture()
def system(db):
    """An empty system."""
    return db.system


@pytest.fixture()
def configuration(system):
    """An empty system."""
    return system.configuration


@pytest.fixture()
def atoms(configuration):
    """An empty atoms table."""
    return configuration.atoms


@pytest.fixture()
def bonds(configuration):
    """An empty bonds table."""
    return configuration.bonds


@pytest.fixture()
def two_dbs():
    """Create two different dbs."""
    db1 = SystemDB(filename="file:seamm_db1?mode=memory&cache=shared")
    system1 = db1.create_system(name="default")
    system1.create_configuration(name="default")

    db2 = SystemDB(filename="file:seamm_db2?mode=memory&cache=shared")
    system2 = db2.create_system(name="default")
    system2.create_configuration(name="default")

    yield db1, db2

    db1.close()
    db2.close()


@pytest.fixture()
def two_systems(two_dbs):
    """Two empty systems."""
    db1, db2 = two_dbs
    return db1.system, db2.system


@pytest.fixture()
def two_configurations(two_systems):
    """Two empty configuration."""
    system1, system2 = two_systems
    return system1.configuration, system2.configuration


@pytest.fixture()
def simple_table(db):
    return mk_table(db)


@pytest.fixture()
def two_tables(db):
    return mk_table(db, "table1"), mk_table(db, "table2")


@pytest.fixture()
def system_with_two_tables(db):
    mk_table(db, "table1")
    mk_table(db, "table2")
    return db.system


@pytest.fixture()
def db_with_two_tables(db):
    mk_table(db, "table1")
    mk_table(db, "table2")
    return db


@pytest.fixture()
def two_tables_in_two_dbs(two_dbs):
    db1, db2 = two_dbs
    return mk_table(db1, "table1"), mk_table(db2, "table2")


@pytest.fixture()
def old_templates(system):
    """A system with a template for water."""
    templates = system["template"]
    atoms = system["templateatom"]
    atoms.add_attribute("name", "str", default="")
    bonds = system["templatebond"]

    # TIP3P
    r0 = 0.9572
    theta0 = 104.52

    # H locations are ±x, 0, z
    x = r0 * math.sin(math.radians(theta0 / 2))
    z = r0 * math.cos(math.radians(theta0 / 2))

    X = [0.0, x, -x]
    Y = [0.0, 0.0, 0.0]
    Z = [0.0, z, z]

    atno = [8, 1, 1]
    name = ["O", "H1", "H2"]
    i_atom = [0, 0]
    j_atom = [1, 2]

    tid = templates.append(name="H2O", type="molecule")[0]
    templates.current_template = tid

    ids = atoms.append(atno=atno, x=X, y=Y, z=Z, name=name)

    i = [ids[x] for x in i_atom]
    j = [ids[x] for x in j_atom]

    bonds.append(i=j, j=i)  # flipped on purpose so code orders.

    # Acetic acid

    # yapf: disable
    #       C       H        H        H        C        =O      O        H
    X = [ 1.0797, 0.5782,  0.7209,  0.7052,  0.5713, -0.1323, 0.9757,  2.1724]  # noqa: E221, E501, E201
    Y = [ 0.0181, 3.1376, -0.6736, -0.3143,  1.3899,  1.7142, 2.2970,  0.0161]  # noqa: E221, E501, E201
    Z = [-0.0184, 0.2813, -0.7859,  0.9529, -0.3161, -1.2568, 0.5919, -0.0306]  # noqa: E221, E501
    symbol = ['C', 'H', 'H', 'H', 'C', 'O', 'O', 'H']
    name = ['C1', 'H1', 'H2', 'H3', 'C', 'O', 'OH', 'H']

    #       C  H  H  H  C =O  O  H
    i_atom = [0, 0, 0, 0, 4, 4, 6]
    j_atom = [1, 2, 3, 4, 5, 6, 7]
    order =  [1, 1, 1, 1, 2, 1, 1]  # noqa: E222
    # yapf: enable

    tid = templates.append(name="acetic acid", type="molecule")[0]
    templates.current_template = tid

    ids = atoms.append(symbol=symbol, x=X, y=Y, z=Z, name=name)

    i = [ids[x] for x in i_atom]
    j = [ids[x] for x in j_atom]

    bonds.append(i=i, j=j, bondorder=order)

    return system


@pytest.fixture()
def AceticAcid(configuration):
    """An configuration object for an acetic acid molecule"""
    # yapf: disable
    #       C       H        H        H        C        =O      O        H
    x = [ 1.0797, 0.5782,  0.7209,  0.7052,  0.5713, -0.1323, 0.9757,  2.1724]  # noqa: E221, E501, E201
    y = [ 0.0181, 3.1376, -0.6736, -0.3143,  1.3899,  1.7142, 2.2970,  0.0161]  # noqa: E221, E501, E201
    z = [-0.0184, 0.2813, -0.7859,  0.9529, -0.3161, -1.2568, 0.5919, -0.0306]  # noqa: E221, E501
    atno = [6, 1, 1, 1, 6, 8, 8, 1]  # noqa: E221

    #       C  H  H  H  C =O  O  H
    i_atom = [0, 0, 0, 0, 4, 4, 6]
    j_atom = [1, 2, 3, 4, 5, 6, 7]
    order =  [1, 1, 1, 1, 2, 1, 1]  # noqa: E222
    # yapf: enable

    configuration.system.name = "acetic acid"
    configuration.name = "acetic acid"
    ids = configuration.atoms.append(x=x, y=y, z=z, atno=atno)

    i = [ids[x] for x in i_atom]
    j = [ids[x] for x in j_atom]

    configuration.bonds.append(i=i, j=j, bondorder=order)

    return configuration


@pytest.fixture
def disordered(AceticAcid):
    """Two acetic acid molecules with different atom order"""
    # yapf: disable
    #       C       H        H        H        C        =O      O        H
    x = [ 1.0797, 0.5782,  0.7209,  0.7052,  0.5713, -0.1323, 0.9757,  2.1724]  # noqa: E221, E501, E201
    y = [ 0.0181, 3.1376, -0.6736, -0.3143,  1.3899,  1.7142, 2.2970,  0.0161]  # noqa: E221, E501, E201
    z = [-0.0184, 0.2813, -0.7859,  0.9529, -0.3161, -1.2568, 0.5919, -0.0306]  # noqa: E221, E501
    atno = [6, 1, 1, 1, 6, 8, 8, 1]  # noqa: E221

    #       C  H  H  H  C =O  O  H
    i_atom = [0, 0, 0, 0, 4, 4, 6]
    j_atom = [1, 2, 3, 4, 5, 6, 7]
    order =  [1, 1, 1, 1, 2, 1, 1]  # noqa: E222
    # yapf: enable

    configuration = AceticAcid
    ids = configuration.atoms.append(
        x=[*reversed(x)], y=[*reversed(y)], z=[*reversed(z)], atno=[*reversed(atno)]
    )

    i = [ids[7 - x] for x in i_atom]
    j = [ids[7 - x] for x in j_atom]

    configuration.bonds.append(i=i, j=j, bondorder=order)

    return configuration


@pytest.fixture()
def vanadium(configuration):
    """BCC vanadium crystal, without symmetry."""
    configuration.name = "BCC Vanadium"
    configuration.periodicity = 3
    configuration.coordinate_system = "fractional"
    configuration.cell.parameters = [3.03, 3.03, 3.03, 90, 90, 90]
    configuration.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol="V")
    return configuration


@pytest.fixture()
def copper(configuration):
    """FCC copper crystal, without symmetry."""
    x = [0.0, 0.5, 0.5, 0.0]
    y = [0.0, 0.5, 0.0, 0.5]
    z = [0.0, 0.0, 0.5, 0.5]
    configuration.name = "FCC Copper"
    configuration.periodicity = 3
    configuration.coordinate_system = "fractional"
    configuration.cell.parameters = (3.61491, 3.61491, 3.61491, 90, 90, 90)
    configuration.atoms.append(x=x, y=y, z=z, symbol=["Cu"])
    return configuration


@pytest.fixture()
def CH3COOH_3H2O(AceticAcid):
    """Configuration with acetic acid and 3 water molecules"""
    configuration = AceticAcid

    # TIP3P
    r0 = 0.9572
    theta0 = 104.52

    # H locations are ±x, 0, z
    x = r0 * math.sin(math.radians(theta0 / 2))
    z = r0 * math.cos(math.radians(theta0 / 2))

    X = [0.0, x, -x]
    Z = [0.0, z, z]

    atno = [8, 1, 1]
    i_atom = [0, 0]
    j_atom = [1, 2]

    configuration.name = "acetic acid with 3 waters"

    for no in range(1, 4):
        ids = configuration.atoms.append(x=X, y=no * 5.0, z=Z, atno=atno)

        i = [ids[x] for x in i_atom]
        j = [ids[x] for x in j_atom]

        configuration.bonds.append(i=i, j=j)

    return configuration


# Fixtures for working with the files in data/


@pytest.fixture()
def full_db():
    """Create a system database with several systems."""
    db = SystemDB(filename="file:seamm_db?mode=memory&cache=shared")
    for filename in ["acy.mmcif"]:
        with open(data_path / filename, "r") as fd:
            text = fd.read()
        system = db.add_system(name=Path(filename).stem)
        configuration = system.create_configuration(name="default")
        configuration.from_mmcif_text(text)

    yield db

    db.close()


@pytest.fixture(scope="session")
def amino_acids():
    """Create a system database with 20 amino acids, each as a system."""
    db = SystemDB(filename="file:amino_acids_db?mode=memory&cache=shared")
    db.read_cif_file(data_path / "aminoacids.mmcif")

    yield db

    db.close()


@pytest.fixture(scope="session")
def aa_templates():
    """Create a system database with 20 amino acids, each as a template."""
    db = SystemDB(filename="file:aa_templates_db?mode=memory&cache=shared")
    db.read_cif_file(data_path / "aminoacids.mmcif")
    templates = db.templates
    for system in db.systems:
        templates.create(
            name=system.name,
            category="amino acid",
            configuration=system.configuration.id,
        )

    yield templates

    db.close()


@pytest.fixture()
def gly(aa_templates):
    """The template for glycine."""
    return aa_templates.get(8)


@pytest.fixture()
def simple_templates(CH3COOH_3H2O):
    """Simple templates for water and acetic acid, plus solvated acetic acid."""
    db = CH3COOH_3H2O.system_db
    templates = db.templates

    db.read_cif_file(data_path / "acy.mmcif")[0]
    db.read_cif_file(data_path / "hoh.mmcif")[0]

    templates.create(name="acy", category="molecule")
    templates.create(name="hoh", category="molecule")

    return db


@pytest.fixture()
def full_templates(CH3COOH_3H2O):
    """Full templates for water and acetic acid, plus solvated acetic acid."""
    db = CH3COOH_3H2O.system_db
    templates = db.templates
    system = CH3COOH_3H2O.system
    system.name = "acetic acid"

    acy_sys = db.read_cif_file(data_path / "acy.mmcif")[0]
    acy_conf = db.get_system(acy_sys).configuration
    acy_cid = acy_conf.id
    hoh_sys = db.read_cif_file(data_path / "hoh.mmcif")[0]
    hoh_conf = db.get_system(hoh_sys).configuration
    hoh_cid = hoh_conf.id

    templates.create(name="acy", category="molecule", configuration=acy_cid)
    templates.create(name="hoh", category="molecule", configuration=hoh_cid)

    return db


@pytest.fixture()
def properties(system):
    """Lots of configurations with many properties."""
    properties = system.system_db.properties

    # Create systems
    cids = []
    for i in range(1000):
        configuration = system.create_configuration(name=str(i))
        cids.append(configuration.id)

    # And properties
    for i in range(10):
        fid = properties.add(f"float_{i}", "float")
        iid = properties.add(f"int_{i}", "int")
        sid = properties.add(f"str_{i}", "str")

        fval = -10.0 + i
        ival = -500 + 100 * i
        for cid in cids:
            properties.put(cid, fid, fval)
            properties.put(cid, iid, ival)
            properties.put(cid, sid, f"string {ival}")
            fval += 0.01
            ival += 1

    return properties



# -*- coding: utf-8 -*-

"""Interface to RDKit."""

import logging

try:
    from rdkit import Chem
except ModuleNotFoundError:
    print(
        "Please install rdkit using conda:\n" "     conda install -c conda-forge rdkit"
    )
    raise

logger = logging.getLogger(__name__)


class RDKitMixin:
    """A mixin for handling RDKit via its Python interface."""

    def to_RDKMol(self):
        """Return an RDKMol object for the configuration, template, or subset."""
        index = {}
        indices = []
        rdk_mol = Chem.RWMol()
        for atno, _id in zip(self.atoms.atomic_numbers, self.atoms.ids):
            idx = rdk_mol.AddAtom(Chem.Atom(atno))
            index[_id] = idx
            indices.append(idx)

        bond_types = {
            1: Chem.BondType.SINGLE,
            2: Chem.BondType.DOUBLE,
            3: Chem.BondType.TRIPLE,
        }
        for row in self.bonds.bonds():
            rdk_mol.AddBond(
                index[row["i"]], index[row["j"]], bond_types[row["bondorder"]]
            )

        natom = self.atoms.n_atoms
        conf = Chem.Conformer(natom)
        for idx, xyz in zip(indices, self.atoms.coordinates):
            conf.SetAtomPosition(idx, xyz)

        rdk_mol.AddConformer(conf)
        Chem.rdmolops.SanitizeMol(rdk_mol)

        return rdk_mol

    def from_RDKMol(self, rdk_mol):
        """Transform an RDKit molecule into the current object."""

        atnos = []
        for rdk_atom in rdk_mol.GetAtoms():
            atnos.append(rdk_atom.GetAtomicNum())
            logger.debug(f"atom {atnos}")

        # TODO: Generalize to handling multiple conformers
        # in a rdk_mol object, if necessary
        Xs = []
        Ys = []
        Zs = []
        for rdk_conf in rdk_mol.GetConformers():
            for atom_coords in rdk_conf.GetPositions():
                Xs.append(atom_coords[0])
                Ys.append(atom_coords[1])
                Zs.append(atom_coords[2])
                logger.debug(f"{atom_coords[0]} {atom_coords[1]} {atom_coords[2]}")

        Is = []
        Js = []
        BondOrders = []
        bond_types = {
            Chem.BondType.SINGLE: 1,
            Chem.BondType.DOUBLE: 2,
            Chem.BondType.TRIPLE: 3,
        }
        for rdk_bond in rdk_mol.GetBonds():
            i = rdk_bond.GetBeginAtom().GetIdx()
            j = rdk_bond.GetEndAtom().GetIdx()
            bondorder = bond_types[rdk_bond.GetBondType()]
            Is.append(i)
            Js.append(j)
            BondOrders.append(bondorder)
            logger.debug(f"bond {i} - {j} {bondorder}")

        self.clear()
        ids = self.atoms.append(x=Xs, y=Ys, z=Zs, atno=atnos)
        i = [ids[x - 1] for x in Is]
        j = [ids[x - 1] for x in Js]
        self.bonds.append(i=i, j=j, bondorder=BondOrders)

        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import pprint  # noqa: F401
import pytest  # noqa: F401

import molsystem  # noqa: F401

"""Tests for `molsystem` package."""

x = [1.0, 2.0, 3.0]
y = [4.0, 5.0, 6.0]
z = [7.0, 8.0, 9.0]
atno = [8, 1, 1]


def test_construction(configuration):
    """Simplest test that we can make an Atoms object"""
    atoms = configuration.atoms
    assert str(type(atoms)) == "<class 'molsystem.atoms._Atoms'>"


def test_keys(atoms):
    """Test the default keys in an Atoms object"""
    result = ["atno", "configuration", "id", "x", "y", "z"]

    assert sorted([*atoms.keys()]) == result


def test_n_atoms_empty(atoms):
    """Test how many atoms an empty object has"""
    assert atoms.n_atoms == 0


def test_append_one(atoms):
    """Test adding one atom"""
    with atoms as tmp:
        tmp.append(x=1.0, y=2.0, z=3.0, atno=[6])
    assert atoms.n_atoms == 1


def test_append_several(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms.n_atoms == 3


def test_append_several_scalar(atoms):
    """Test adding several atoms with some scalar values"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=0.0, atno=6)
    assert atoms["x"] == x
    assert atoms["y"] == y
    assert atoms["z"] == [0.0, 0.0, 0.0]
    assert atoms["atno"] == [6, 6, 6]


def test_append_using_default(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z)

    assert atoms.n_atoms == 3
    assert atoms["atno"] == [None, None, None]


def test_append_error(atoms):
    """Test adding atoms with an error"""
    with pytest.raises(KeyError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno, bad=99)
    assert atoms.n_atoms == 0
    assert atoms.configuration.version == 0
    assert str(e.value) == "'\"bad\" is not an attribute of the atoms.'"


def test_add_attribute(atoms):
    """Test adding an attribute"""
    result = ["atno", "configuration", "id", "name", "x", "y", "z"]
    with atoms as tmp:
        tmp.add_attribute("name")
    assert sorted([*atoms.keys()]) == result


def test_add_duplicate_attribute(atoms):
    """Test duplicate adding an attribute"""
    result = ["atno", "configuration", "id", "name", "x", "y", "z"]
    with atoms as tmp:
        tmp.add_attribute("name")
    with pytest.raises(RuntimeError) as e:
        with atoms as tmp:
            ver = tmp.configuration.version
            tmp.add_attribute("name")
    assert sorted([*atoms.keys()]) == result
    assert ver == 1
    assert atoms.configuration.version == 1
    assert str(e.value) == "_Table attribute 'name' is already defined!"


def test_add_coordinates_attribute(atoms):
    """Test adding an attribute"""
    result = ["atno", "configuration", "id", "spin", "x", "y", "z"]
    with atoms as tmp:
        tmp.add_attribute("spin", configuration_dependent=True)
    assert sorted([*atoms.keys()]) == result
    del atoms["spin"]
    result.remove("spin")
    assert sorted([*atoms.keys()]) == result


def test_add_attribute_with_values(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("name", values=["H1", "O", "H2"])
    assert np.array_equal(atoms["name"], ["H1", "O", "H2"])


def test_add_attribute_with_one_value(atoms):
    """Test adding several atoms"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("name", values=["H1"])
    assert np.array_equal(atoms["name"], ["H1", "H1", "H1"])


def test_add_attribute_with_one_value_not_list(atoms):
    """Test adding an attribute using a scalar value"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("spin", coltype="int", default=0, values=1)
    assert atoms["spin"] == [1, 1, 1]


def test_add_attribute_with_wrong_number_of_values(atoms):
    """Test adding an attribute using a the wrong number of values"""
    with pytest.raises(IndexError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno)
            tmp.add_attribute("spin", coltype="int", default=0, values=[1, 2])
    assert atoms.n_atoms == 0
    assert str(e.value) == (
        "The number of values given, "
        '2, must be either 1, or the number of rows in "main"."atom": 3'
    )


def test_get_attribute_by_index(atoms):
    """Get a single value of an attribute by index"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms["atno"][1] == 1


def test_get_attribute_by_slice(atoms):
    """Get several values using a slice"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms["atno"][::2] == [8, 1]


def test_getting_x(atoms):
    """Test getting the x column of coordinates"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert atoms["x"] == [1.0, 2.0, 3.0]


def test_contains(atoms):
    """Test the __contains__ or 'in' functionalty"""
    assert "x" in atoms


def test_not_in(atoms):
    """Test the __contains__ or 'in' functionalty"""
    assert "abc" not in atoms


def test_deleting_column(atoms):
    """Test deleting a column"""
    with atoms as tmp:
        del tmp["atno"]
    assert sorted([*atoms.keys()]) == ["configuration", "id", "x", "y", "z"]


def test_set_column(atoms):
    """Test setting a column using a scalar"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms as tmp:
        tmp["atno"] = 10

    assert atoms.n_atoms == 3
    assert atoms.configuration.version == 2
    assert atoms["atno"] == [10, 10, 10]


def test_set_column_with_array(atoms):
    """Test setting a column using an array"""
    values = [10.0, 11.0, 12.0]

    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms as tmp:
        tmp["x"] = values

    assert atoms.n_atoms == 3
    assert atoms.configuration.version == 2
    assert atoms["x"] == values


def test_append_error_no_coordinates(atoms):
    """Test adding atoms without coordinates, raising an error"""
    with atoms as tmp:
        tmp.append(y=y, z=z, atno=atno)

    assert atoms.n_atoms == 3
    assert atoms["x"] == [None, None, None]


def test_append_error_invalid_column(atoms):
    """Test adding atoms without coordinates, raising an error"""
    with pytest.raises(KeyError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno, junk=99)
    assert atoms.n_atoms == 0
    assert str(e.value) == "'\"junk\" is not an attribute of the atoms.'"


def test_append_error_invalid_length(atoms):
    """Test adding atoms without coordinates, raising an error"""
    with pytest.raises(IndexError) as e:
        with atoms as tmp:
            tmp.append(x=x, y=y, z=z, atno=[3, 4])
    assert atoms.n_atoms == 0
    assert str(e.value) == (
        'key "atno" has the wrong number of values, '
        "2. Should be 1 or the number of atoms (3)."
    )


def test_add_attribute_with_no_default(atoms):
    """Test adding an attribute with no default, then several atoms"""
    with atoms as tmp:
        tmp.add_attribute("new", coltype="float")
        tmp.append(x=x, y=y, z=z, atno=atno, new=[-1, -2, -3])

    assert atoms.n_atoms == 3
    assert atoms["new"] == [-1, -2, -3]


def test_equality(two_configurations):
    """Test whether two Atoms objects are equal"""
    configuration1, configuration2 = two_configurations
    atoms1 = configuration1.atoms
    atoms2 = configuration2.atoms

    with atoms1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms2 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    assert atoms1.n_atoms == 3
    assert atoms1 == atoms2


def test_inequality(two_configurations):
    """Test whether two Atoms objects are equal"""
    configuration1, configuration2 = two_configurations
    atoms1 = configuration1.atoms
    atoms2 = configuration2.atoms

    with atoms1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with atoms2 as tmp:
        atno2 = list(atno)
        atno2[2] = 13
        tmp.append(x=x, y=y, z=z, atno=atno2)

    assert atoms1.n_atoms == 3
    assert atoms1 != atoms2


def test_str(atoms):
    """Test string representation of atoms object"""
    string_rep = """\
   atno  configuration    x    y    z
1     8              1  1.0  4.0  7.0
2     1              1  2.0  5.0  8.0
3     1              1  3.0  6.0  9.0"""  # noqa: W291

    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    if str(atoms) != string_rep:
        print(str(atoms))

    assert str(atoms) == string_rep


def test_repr(atoms):
    """Test representation of atoms object"""
    string_rep = """\
   atno  configuration    x    y    z
1     8              1  1.0  4.0  7.0
2     1              1  2.0  5.0  8.0
3     1              1  3.0  6.0  9.0"""  # noqa: W291

    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    if repr(atoms) != string_rep:
        print(repr(atoms))

    assert repr(atoms) == string_rep


def test_equals(atoms):
    """Test shallow copy"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    atoms2 = atoms

    assert atoms == atoms2


def test_equals_with_change(atoms):
    """Test (shallow) copy and changing copy"""
    with atoms as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    atoms2 = atoms
    atoms2["atno"][0] = 22

    assert atoms == atoms2


def test_coordinates(AceticAcid):
    """Test getting coordinates."""
    configuration = AceticAcid

    xs = [1.0797, 0.5782, 0.7209, 0.7052, 0.5713, -0.1323, 0.9757, 2.1724]
    ys = [0.0181, 3.1376, -0.6736, -0.3143, 1.3899, 1.7142, 2.2970, 0.0161]
    zs = [-0.0184, 0.2813, -0.7859, 0.9529, -0.3161, -1.2568, 0.5919, -0.0306]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    xyz = configuration.atoms.coordinates
    assert np.allclose(xyz, xyz0)


def test_set_coordinates(AceticAcid):
    """Test setting coordinates."""
    configuration = AceticAcid

    xs = [1.08, 0.58, 0.72, 0.71, 0.57, -0.13, 0.98, 2.17]
    ys = [0.02, 3.14, -0.67, -0.31, 1.39, 1.71, 2.30, 0.02]
    zs = [-0.02, 0.28, -0.79, 0.95, -0.32, -1.26, 0.59, -0.03]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    configuration.atoms.set_coordinates(xyz0)

    xyz = configuration.atoms.coordinates
    assert np.allclose(xyz, xyz0)


def test_selected_atoms(AceticAcid):
    """Test getting the number of selected atoms."""
    configuration = AceticAcid

    assert configuration.atoms.get_n_atoms("atno", "==", 6) == 2
    x = []
    for row in configuration.atoms.atoms("atno", "==", 6):
        x.append(row["x"])
    assert x == [1.0797, 0.5713]


def test_periodic_coordinates(vanadium):
    """Test getting coordinates."""
    configuration = vanadium

    xs = [0.0, 0.5]
    ys = [0.0, 0.5]
    zs = [0.0, 0.5]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    xyz = configuration.atoms.coordinates
    assert np.allclose(xyz, xyz0)


def test_periodic_coordinates_cartesians(vanadium):
    """Test getting coordinates in Cartesian coordinates."""
    configuration = vanadium

    xs = [0.0, 1.515]
    ys = [0.0, 1.515]
    zs = [0.0, 1.515]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    xyz = configuration.atoms.get_coordinates(fractionals=False)
    assert np.allclose(xyz, xyz0)


def test_set_periodic_coordinates(vanadium):
    """Test setting coordinates."""
    configuration = vanadium

    xs = [0.1, 0.6]
    ys = [0.1, 0.6]
    zs = [0.1, 0.6]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    configuration.atoms.set_coordinates(xyz0)

    xyz = configuration.atoms.coordinates
    assert np.allclose(xyz, xyz0)


def test_set_periodic_coordinates_cartesians(vanadium):
    """Test setting coordinates in Cartesian coordinates."""
    configuration = vanadium

    xs = [0.303, 1.818]
    ys = [0.303, 1.818]
    zs = [0.303, 1.818]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    configuration.atoms.set_coordinates(xyz0, fractionals=False)

    xyz = configuration.atoms.get_coordinates(fractionals=False)
    assert np.allclose(xyz, xyz0)

    xs = [0.1, 0.6]
    ys = [0.1, 0.6]
    zs = [0.1, 0.6]
    xyz0 = [[x, y, z] for x, y, z in zip(xs, ys, zs)]

    xyz = configuration.atoms.coordinates
    assert np.allclose(xyz, xyz0)


def test_remove_atoms(copper):
    """Test removing one atom from FCC copper"""
    configuration = copper

    assert configuration.atoms.n_atoms == 4

    ids = configuration.atoms.ids
    first = [ids[0]]

    with configuration.atoms as tmp:
        tmp.delete(atoms=first)

    assert configuration.atoms.n_atoms == 3

    xyz = configuration.atoms.coordinates

    assert xyz == [[0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401

import molsystem  # noqa: F401

"""Tests for templates in the `molsystem` package."""

x = [1.0, 2.0, 3.0]
y = [4.0, 5.0, 6.0]
z = [7.0, 8.0, 9.0]
atno = [8, 1, 1]


def test_construction(db):
    """Simplest test that we can make a new templates object"""
    templates = db.templates
    assert str(type(templates)) == "<class 'molsystem.templates._Templates'>"


def test_create(db):
    """Test that we can make a new template"""
    templates = db.templates
    template = templates.create(name="H2O", category="molecule")
    assert template.id == 1


def test_create_many(db):
    """Test that we can make several templates."""
    templates = db.templates
    names = ["H2O", "H2", "O2"]
    new = templates.create_many(name=names, category="molecule")
    assert [x.id for x in new] == [1, 2, 3]
    assert templates.categories == ["molecule"]
    assert templates.names("molecule") == sorted(names)


def test_making_full_templates(amino_acids):
    """Create full templates."""
    answer = [
        "ALA",
        "ARG",
        "ASN",
        "ASP",
        "CYS",
        "GLN",
        "GLU",
        "GLY",
        "HIS",
        "ILE",
        "LEU",
        "LYS",
        "MET",
        "PHE",
        "PRO",
        "SER",
        "THR",
        "TRP",
        "TYR",
        "VAL",
    ]
    templates = amino_acids.templates
    for system in amino_acids.systems:
        templates.create(
            name=system.name,
            category="amino acid",
            configuration=system.configuration.id,
        )
    assert amino_acids.n_templates == 20
    assert templates.names("amino acid") == answer


def test_get_template(aa_templates):
    """Test getting a template by index and name."""
    template = aa_templates.get(8)
    assert template.name == "GLY"
    template = aa_templates.get("GLU", category="amino acid")
    assert template.id == 6


def test_template(gly):
    """Test the gly template."""
    assert gly.n_atoms == 10
    assert gly.n_bonds == 9


def test_template_coordinate_system(gly):
    """Test the gly template coordinate system."""
    assert gly.coordinate_system == "Cartesian"


def test_template_formula(gly):
    """Test the gly template formula."""
    assert gly.formula == ("C2 H5 N O2", "C2 H5 N O2", 1)


def test_template_mass(gly):
    """Test the gly template mass."""
    assert abs(gly.mass - 75.067) <= 0.001


def test_template_periodicity(gly):
    """Test the gly template periodicity."""
    assert gly.periodicity == 0


def test_template_atoms(gly):
    """Test the atoms for the template."""
    answer = ["N", "CA", "C", "O", "OXT", "H", "H2", "HA2", "HA3", "HXT"]
    names = gly.atoms.get_column_data("name")
    if names != answer:
        print(names)
    assert names == answer


def test_atoms_data(gly):
    """Test the data for the atoms for the template."""
    answer = {
        "id": [127, 128, 129, 130, 131, 132, 133, 134, 135, 136],
        "atno": [7, 6, 6, 8, 8, 1, 1, 1, 1, 1],
        "name": ["N", "CA", "C", "O", "OXT", "H", "H2", "HA2", "HA3", "HXT"],
        "formal_charge": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        "_chem_comp_atom.alt_atom_id": [
            "N",
            "CA",
            "C",
            "O",
            "OXT",
            "H",
            "HN2",
            "HA1",
            "HA2",
            "HXT",
        ],
        "_chem_comp_atom.pdbx_align": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        "_chem_comp_atom.pdbx_aromatic_flag": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        "_chem_comp_atom.pdbx_leaving_atom_flag": [1, 1, 1, 1, 0, 1, 0, 1, 1, 0],
        "_chem_comp_atom.pdbx_stereo_config": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        "_chem_comp_atom.pdbx_component_atom_id": [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        "configuration": [8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
        "x": [
            25.463,
            25.329,
            26.081,
            27.024,
            25.702,
            25.494,
            26.307,
            24.27,
            25.731,
            26.236,
        ],
        "y": [
            35.609,
            37.024,
            37.335,
            36.627,
            38.256,
            35.15,
            35.421,
            37.305,
            37.59,
            38.3,
        ],
        "z": [
            47.047,
            46.85,
            45.572,
            45.222,
            44.874,
            46.159,
            47.549,
            46.757,
            47.703,
            44.09,
        ],
    }
    for column in gly.atoms:
        result = gly.atoms.get_column_data(column)
        if column not in answer:
            print(f"'{column}': {result}")
        assert column in answer
        if result != answer[column]:
            print(f"'{column}': {result}")
        assert result == answer[column]


def test_template_bonds(gly):
    """Test the bonds for the template."""
    answer = [1, 1, 1, 1, 1, 1, 2, 1, 1]
    bondorders = gly.bonds.get_column_data("bondorder")
    if bondorders != answer:
        print(bondorders)
    assert bondorders == answer


def test_print_template_bonds(gly):
    """Test the string reprentation of the bonds in the template."""
    answer = """\
       i    j  bondorder _chem_comp_bond.comp_id
120  127  128          1                     GLY
121  127  132          1                     GLY
122  127  133          1                     GLY
123  128  129          1                     GLY
124  128  134          1                     GLY
125  128  135          1                     GLY
126  129  130          2                     GLY
127  129  131          1                     GLY
128  131  136          1                     GLY"""
    result = str(gly.bonds)
    if result != answer:
        print(result)
    assert result == answer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import pprint

import numpy as np
import pytest  # noqa: F401

import molsystem  # noqa: F401

"""Tests for the Table class."""

x = [1.0, 2.0, 3.0]
y = [4.0, 5.0, 6.0]
z = [7.0, 8.0, 9.0]
atno = [8, 1, 1]


def test_construction(simple_table):
    """Simplest test that we can make an Table object"""
    assert str(type(simple_table)) == "<class 'molsystem.table._Table'>"


def test_keys(simple_table):
    """Test the default keys in an Table object"""
    assert sorted([*simple_table.keys()]) == ["atno", "x", "y", "z"]


def test_n_rows_empty(simple_table):
    """Test how many rows an empty object has"""
    assert simple_table.n_rows == 0


def test_append_one(simple_table):
    """Test adding one row"""
    with simple_table as tmp:
        tmp.append(x=1.0, y=2.0, z=3.0, atno=[6])
    assert simple_table.n_rows == 1


def test_append_several(simple_table):
    """Test adding several rows"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert simple_table.n_rows == 3


def test_append_several_scalar(simple_table):
    """Test adding several simple_table with some scalar values"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=0.0, atno=6)
    assert (
        simple_table["x"] == x
        and simple_table["y"] == y
        and simple_table["z"] == [0.0, 0.0, 0.0]
        and simple_table["atno"] == [6, 6, 6]
    )


def test_append_several_scalar_first(simple_table):
    """Test adding several simple_table with a scalar value first"""
    with simple_table as tmp:
        tmp.append(atno=6, x=x, y=y, z=0.0)
    assert (
        simple_table["x"] == x
        and simple_table["y"] == y
        and simple_table["z"] == [0.0, 0.0, 0.0]
        and simple_table["atno"] == [6, 6, 6]
    )


def test_append_using_default(simple_table):
    """Test adding several rows using default values"""

    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z)

    assert simple_table.n_rows == 3 and simple_table["atno"] == [-1, -1, -1]


def test_append_ndarrays_3x100(simple_table):
    """Test adding 3 rows 100 times, to force allocation"""
    xaa = np.array(x)
    yaa = np.array(y)
    zaa = np.array(z)
    atnoaa = np.array(atno)
    with simple_table as tmp:
        for i in range(0, 100):
            tmp.append(
                x=xaa.tolist(), y=yaa.tolist(), z=zaa.tolist(), atno=atnoaa.tolist()
            )
            xaa += 10.0
            yaa += 5.0
            zaa -= 3.0
    assert simple_table.n_rows == 300


def test_append_ndarrays_3x10_one_by_one(simple_table):
    """Test adding 3 rows 10 times separately, checking changes"""
    xaa = np.array(x)
    yaa = np.array(y)
    zaa = np.array(z)
    atnoaa = np.array(atno)
    for i in range(0, 10):
        with simple_table as tmp:
            tmp.append(
                x=xaa.tolist(), y=yaa.tolist(), z=zaa.tolist(), atno=atnoaa.tolist()
            )
            xaa += 10.0
            yaa += 5.0
            zaa -= 3.0
    assert simple_table.n_rows == 30


def test_append_error(simple_table):
    """Test adding rows with an error"""
    try:
        with simple_table as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno)
            raise RuntimeError()
    except RuntimeError:
        pass
    assert simple_table.n_rows == 0


def test_add_attribute(simple_table):
    """Test adding an attribute"""
    with simple_table as tmp:
        tmp.add_attribute("name")
    assert sorted([*simple_table.keys()]) == ["atno", "name", "x", "y", "z"]


def test_add_duplicate_attribute(simple_table):
    """Test duplicate adding an attribute"""
    with simple_table as tmp:
        tmp.add_attribute("name")
    try:
        with simple_table as tmp:
            tmp.add_attribute("name")
    except RuntimeError as e:
        err = str(e)
    assert sorted([*simple_table.keys()]) == ["atno", "name", "x", "y", "z"]
    assert err == "_Table attribute 'name' is already defined!"


def test_add_attribute_with_values(simple_table):
    """Test adding an attribute with values"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("name", values=["H1", "O", "H2"])
    assert simple_table["name"] == ["H1", "O", "H2"]


def test_add_attribute_with_one_value(simple_table):
    """Test adding an attribute with a list length 1"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("name", values=["H1"])
    assert simple_table["name"] == ["H1", "H1", "H1"]


def test_add_attribute_with_one_value_not_list(simple_table):
    """Test adding an attribute using a scalar value"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("spin", coltype="int", default=0, values=1)
    assert simple_table["spin"] == [1, 1, 1]


def test_add_attribute_with_wrong_number_of_values(simple_table):
    """Test adding an attribute using the wrong number of values"""
    try:
        with simple_table as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno)
            tmp.add_attribute("spin", coltype="int", default=0, values=[1, 2])
    except IndexError as e:
        err = str(e)

    assert simple_table.n_rows == 0 and err == (
        "The number of values given, 2, must be either 1, or the number of"
        ' rows in "main"."table1": 3'
    )


def test_get_attribute_by_index(simple_table):
    """Get a single value of an attribute by index"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert simple_table["atno"][1] == 1


def test_get_attribute_by_slice(simple_table):
    """Get several values using a slice"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert simple_table["atno"][::2] == [8, 1]


def test_getting_x(simple_table):
    """Test getting the x column of coordinates"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
    assert simple_table["x"] == [1.0, 2.0, 3.0]


def test_contains(simple_table):
    """Test the __contains__ or 'in' functionalty"""
    assert "x" in simple_table


def test_not_in(simple_table):
    """Test the __contains__ or 'in' functionalty"""
    assert "abc" not in simple_table


def test_deleting_column(simple_table):
    """Test deleting a column"""
    with simple_table as tmp:
        del tmp["atno"]
    assert sorted([*simple_table.keys()]) == ["x", "y", "z"]


def test_set_column(simple_table):
    """Test setting a column using a scalar"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with simple_table as tmp:
        tmp["atno"] = 10

    assert simple_table.n_rows == 3
    assert simple_table["atno"] == [10, 10, 10]


def test_set_column_with_array(simple_table):
    """Test setting a column using an array"""
    values = [10.0, 11.0, 12.0]

    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with simple_table as tmp:
        tmp["x"] = values

    assert simple_table.n_rows == 3
    assert simple_table["x"] == values


def test_append_error_invalid_column(simple_table):
    """Test adding rows with an unknown attribute, raising an error"""
    try:
        with simple_table as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno, junk=3)
    except KeyError as e:
        err = str(e)

    assert (
        simple_table.n_rows == 0
        and err == '\'"junk" is not an attribute of the table "main"."table1"!\''
    )


def test_append_error_invalid_length(simple_table):
    """Test appending with the wrong number of values, raising an error"""
    try:
        with simple_table as tmp:
            tmp.append(x=x, y=y, z=z, atno=atno[0:2])
    except IndexError as e:
        err = str(e)

    assert simple_table.n_rows == 0 and err == (
        'key "atno" has the wrong number of values, '
        '2. Should be 1 or the number of rows in "main"."table1" (3).'
    )


def test_add_attribute_with_no_default(simple_table):
    """Test adding an attribute with no default, then several rows"""
    with simple_table as tmp:
        tmp.add_attribute("new", coltype="float")
        tmp.append(x=x, y=y, z=z, atno=atno, new=[-1, -2, -3])

    assert simple_table.n_rows == 3 and simple_table["new"] == [-1, -2, -3]


def test_add_attribute_with_no_default_error(simple_table):
    """Test adding a NOT NULL attribute with no default."""
    with pytest.raises(ValueError) as e:
        with simple_table as tmp:
            tmp.add_attribute("new", coltype="float", notnull=True)
    assert simple_table.n_rows == 0 and str(e.value) == (
        "Not null attributes must have defaults: new"
    )


def test_equality(two_tables):
    """Test whether two simple_table objects are equal"""
    table1, table2 = two_tables

    with table1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with table2 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    assert table1.n_rows == 3 and table1 == table2


def test_inequality(two_tables):
    """Test whether two simple_table objects are not equal"""
    table1, table2 = two_tables

    with table1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    with table2 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp["atno"][2] = 13

    assert table1.n_rows == 3 and table1 != table2


def test_str(simple_table):
    """Test string representation of simple_table object"""
    string_rep = """\
   atno    x    y    z
1     8  1.0  4.0  7.0
2     1  2.0  5.0  8.0
3     1  3.0  6.0  9.0"""  # noqa: W291

    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    assert str(simple_table) == string_rep


def test_repr(simple_table):
    """Test representation of simple_table object"""
    string_rep = """\
   atno    x    y    z
1     8  1.0  4.0  7.0
2     1  2.0  5.0  8.0
3     1  3.0  6.0  9.0"""  # noqa: W291

    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    assert repr(simple_table) == string_rep


def test_equals(simple_table):
    """Test copy using equals"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    simple_table2 = simple_table

    assert simple_table == simple_table2


def test_equals_with_change(simple_table):
    """Test copy using equals and changing copy"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    simple_table2 = simple_table

    atnos = simple_table2["atno"]
    atnos[2] = 22
    simple_table2["atno"] = atnos

    assert simple_table == simple_table2


def test_copy_with_change(simple_table):
    """Test copy and changing copy"""
    with simple_table as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    simple_table2 = copy.copy(simple_table)

    atnos = simple_table2["atno"]
    atnos[2] = 22
    simple_table2["atno"] = atnos

    assert simple_table == simple_table2


def test_diff(two_tables):
    """Test diffing two tables in the same database."""

    ref1 = {
        "changed": {1: {("atno", 8, 10), ("x", 1.0, 0.0)}},
        "columns in added rows": ["atno", "x", "y", "z"],
        "added": {4: (12, 20.0, 21.0, 22.0)},
        "summary": {"columns changed": {"atno", "x"}, "rows added": 1},
    }

    ref2 = {
        "changed": {1: {("atno", 10, 8), ("x", 0.0, 1.0)}},
        "columns in deleted rows": ["atno", "x", "y", "z"],
        "deleted": {4: (12, 20.0, 21.0, 22.0)},
        "summary": {"columns changed": {"atno", "x"}, "rows deleted": 1},
    }

    table1, table2 = two_tables

    with table1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    x1 = list(x)
    x1[0] = 0.0
    with table2 as tmp:
        tmp.append(x=x1, y=y, z=z, atno=[10, 1, 1])
        tmp.append(x=20.0, y=21.0, z=22.0, atno=12)

    diffs = table2.diff(table1)
    if diffs != ref1:
        pprint.pprint(diffs)
    assert diffs == ref1

    diffs = table1.diff(table2)
    if diffs != ref2:
        pprint.pprint(diffs)
    assert diffs == ref2


def test_diff_two_dbs(two_tables_in_two_dbs):
    """Test diffing two tables in the different databases."""

    ref1 = {
        "changed": {1: {("atno", 8, 10), ("x", 1.0, 0.0)}},
        "columns in added rows": ["atno", "x", "y", "z"],
        "added": {4: (12, 20.0, 21.0, 22.0)},
        "summary": {"columns changed": {"atno", "x"}, "rows added": 1},
    }

    ref2 = {
        "changed": {1: {("atno", 10, 8), ("x", 0.0, 1.0)}},
        "columns in deleted rows": ["atno", "x", "y", "z"],
        "deleted": {4: (12, 20.0, 21.0, 22.0)},
        "summary": {"columns changed": {"atno", "x"}, "rows deleted": 1},
    }

    table1, table2 = two_tables_in_two_dbs

    with table1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)

    x1 = list(x)
    x1[0] = 0.0
    with table2 as tmp:
        tmp.append(x=x1, y=y, z=z, atno=[10, 1, 1])
        tmp.append(x=20.0, y=21.0, z=22.0, atno=12)

    diffs = table2.diff(table1)
    if diffs != ref1:
        pprint.pprint(diffs)
    assert diffs == ref1

    diffs = table1.diff(table2)
    if diffs != ref2:
        pprint.pprint(diffs)
    assert diffs == ref2


def test_diff_add_columns(two_tables_in_two_dbs):
    """Test diffing two tables with different numbers of columns."""

    ref1 = {
        "columns deleted": {"spin"},
        "changed": {1: {("x", 1.0, 0.0), ("atno", 8, 10)}},
        "columns in added rows": ["atno", "x", "y", "z"],
        "added": {4: (12, 20.0, 21.0, 22.0)},
        "summary": {
            "columns changed": {"atno", "x"},
            "columns deleted": {"spin"},
            "rows added": 1,
        },
    }  # yapf: disable

    ref2 = {
        "columns added": {"spin"},
        "changed": {1: {("x", 0.0, 1.0), ("atno", 10, 8)}},
        "columns in deleted rows": ["atno", "x", "y", "z"],
        "deleted": {4: (12, 20.0, 21.0, 22.0)},
        "summary": {
            "columns add": {"spin"},
            "columns changed": {"atno", "x"},
            "rows deleted": 1,
        },
    }  # yapf: disable

    table1, table2 = two_tables_in_two_dbs

    with table1 as tmp:
        tmp.append(x=x, y=y, z=z, atno=atno)
        tmp.add_attribute("spin", coltype="float", default=0.0)

    x1 = list(x)
    x1[0] = 0.0
    with table2 as tmp:
        tmp.append(x=x1, y=y, z=z, atno=[10, 1, 1])
        tmp.append(x=20.0, y=21.0, z=22.0, atno=12)

    diffs = table2.diff(table1)
    if diffs != ref1:
        pprint.pprint(diffs)
    assert diffs == ref1

    diffs = table1.diff(table2)
    if diffs != ref2:
        pprint.pprint(diffs)
    assert diffs == ref2



# -*- coding: utf-8 -*-

import collections.abc
import logging

logger = logging.getLogger(__name__)


class _FrozenColumn(collections.abc.Sequence):
    """A list-like object for holding a column data

    This is a wrapper around a single column in a SQL table.
    """

    def __init__(self, table, column: str, sql=None, where=""):
        self._table = table
        self._column = column
        self._sql = sql
        self._where = where
        self._data = None

        self._initialize()

    def __getitem__(self, index):
        """Allow [] to access the data!"""
        return self._data[index]

    def __len__(self) -> int:
        """The len() command"""
        return len(self._data)

    def __repr__(self) -> str:
        """The string representation of this object"""
        return repr(self._data)

    def __str__(self) -> str:
        """The pretty string representation of this object"""
        return str(self._data)

    def __eq__(self, other):
        """If this is equal to another list."""
        return self._data == other

    @property
    def column(self):
        """A nicely quoted version of the column name"""
        return '"' + self._column + '"'

    def _initialize(self):
        """Get the column data from the underlying table and cache it."""
        self._data = []
        if self._sql is None:
            sql = f"SELECT {self.column} FROM {self._table.table} {self._where}"
        else:
            sql = sql._sql

        for row in self._table.db.execute(sql):
            self._data.append(row[0])

    def equal(self, other, tol=1.0e-6):
        """Check if we are equal to another iterable to within a tolerance.

        Parameters
        ----------
        other : iterable
            The other object to check against
        tol : float = 1.0e-06
            The tolerance for comparing floating point numbers.

        Returns
        -------
        equals : bool
            Boolean indicating whether the two are equal.
        """
        if len(self._data) != len(other):
            return False

        if len(self._data) == 0:
            return True

        if isinstance(self._data[0], float):
            for v1, v2 in zip(self._data, other):
                if abs(v1 - v2) > tol:
                    return False
            return True
        else:
            return self._data == other



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `cell` in the `molsystem` package."""

import pytest  # noqa: F401

from molsystem import Cell


def test_construction(configuration):
    """Simplest test that we can make a Cell object"""
    configuration.periodicity = 3
    cell = configuration.cell
    assert str(type(cell)) == "<class 'molsystem.cell._Cell'>"


def test_construction_error(configuration):
    """Test that we get an error for nonperiodic systems"""
    with pytest.raises(TypeError) as e:
        configuration.cell
    if str(e.value) != "The configuration is not periodic!":
        print(e.value)
    assert str(e.value) == "The configuration is not periodic!"


def test_periodic_system(configuration):
    """Test making a simple periodic system."""
    configuration.periodicity = 3
    configuration.coordinate_system = "fractional"
    configuration.cell.parameters = (3.03, 3.03, 3.03, 90, 90, 90)
    configuration.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol="V")
    assert configuration.atoms.n_atoms == 2
    assert configuration.version == 0


def test_periodic_context(configuration):
    """Test using context."""
    with configuration as tmp:
        tmp.periodicity = 3
        tmp.coordinate_system = "fractional"
        tmp.cell.parameters = (3.03, 3.03, 3.03, 90, 90, 90)
        tmp.atoms.append(x=[0.0, 0.5], y=[0.0, 0.5], z=[0.0, 0.5], symbol="V")
    assert configuration.atoms.n_atoms == 2
    assert configuration.version == 1


def test_cell_object(vanadium):
    """Test getting the cell as a Cell object."""
    cell = vanadium.cell
    assert str(type(cell)) == "<class 'molsystem.cell._Cell'>"
    assert cell == [3.03, 3.03, 3.03, 90.0, 90.0, 90.0]


def test_cell_vectors():
    """Test the simple Cell objects handling of vectors."""
    cell1 = Cell(2.456, 2.456, 6.696, 90.0, 90.0, 120.0)
    vectors = cell1.vectors()
    cell2 = Cell(1, 1, 1, 90, 90, 90)
    cell2.from_vectors(vectors)
    if cell1 != cell2:
        print(f"{cell1=}")
        print(f"{cell2=}")
    assert cell1 == cell2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the RDKit mixin of the class."""

import pprint  # noqa: F401
import rdkit
import rdkit.Chem
import rdkit.Chem.AllChem
import pytest  # noqa: F401


def test_to_RDKMol(configuration):
    """Test creating a RDKMol object from a structure."""
    mol = configuration.to_RDKMol()

    bond_types = {
        rdkit.Chem.rdchem.BondType.SINGLE: 1,
        rdkit.Chem.rdchem.BondType.DOUBLE: 2,
        rdkit.Chem.rdchem.BondType.TRIPLE: 3,
    }
    bond_list = [bond_types[bt.GetBondType()] for bt in mol.GetBonds()]

    assert configuration.n_atoms == mol.GetNumAtoms()
    assert configuration.bonds.bondorders == bond_list


def test_from_RDKMol(configuration):
    """Test creating a structure from an RDKMol object."""
    mol = rdkit.Chem.MolFromSmiles("CC=O")
    mol = rdkit.Chem.rdmolops.AddHs(mol, addCoords=True)

    rdkit.Chem.AllChem.EmbedMolecule(mol)

    configuration.from_RDKMol(mol)

    assert configuration.n_atoms == 7
    assert configuration.bonds.bondorders == [1, 2, 1, 1, 1, 1]
